###############################################################################
# Universal Analytics for Python
# Copyright (c) 2013, Analytics Pros
#
# This project is free software, distributed under the BSD license.
# Analytics Pros offers consulting and integration services if your firm needs
# assistance in strategy, implementation, or auditing existing work.
###############################################################################

#
# reworked version
# migrated to Python 3
# and async/await
#

import urllib
import aiohttp
import asyncio
from urllib.parse import urlencode

import datetime
import hashlib
import logging
import time
import uuid

logger = logging.getLogger(__name__)


def generate_uuid(basedata=None):
    """ Provides a _random_ UUID with no input, or a UUID4-format MD5 checksum of any input data provided """
    if basedata is None:
        return str(uuid.uuid4())
    elif isinstance(basedata, str):
        checksum = hashlib.md5(basedata).hexdigest()
        return '%8s-%4s-%4s-%4s-%12s' % (
            checksum[0:8], checksum[8:12], checksum[12:16], checksum[16:20], checksum[20:32])


class Time(datetime.datetime):
    """ Wrappers and convenience methods for processing various time representations """

    @classmethod
    def from_unix(cls, seconds, milliseconds=0):
        """ Produce a full |datetime.datetime| object from a Unix timestamp """
        base = list(time.gmtime(seconds))[0:6]
        base.append(milliseconds * 1000)  # microseconds
        return cls(*base)

    @classmethod
    def to_unix(cls, timestamp):
        """ Wrapper over time module to produce Unix epoch time as a float """
        if not isinstance(timestamp, datetime.datetime):
            raise TypeError('Time.milliseconds expects a datetime object')
        base = time.mktime(timestamp.timetuple())
        return base

    @classmethod
    def milliseconds_offset(cls, timestamp, now=None):
        """ Offset time (in milliseconds) from a |datetime.datetime| object to now """
        if isinstance(timestamp, (int, float)):
            base = timestamp
        else:
            base = cls.to_unix(timestamp)
            base += (timestamp.microsecond / 1000000)
        if now is None:
            now = time.time()
        return (now - base) * 1000


class HTTPRequest:
    """ URL Construction and request handling abstraction.
        This is not intended to be used outside this module.

        Automates mapping of persistent state (i.e. query parameters)
        onto transcient datasets for each query.
    """

    endpoint = 'https://www.google-analytics.com/collect'

    # Store properties for all requests
    def __init__(self, user_agent=None, *args, **opts):
        self.user_agent = user_agent or 'Bot Story'

    @classmethod
    def fixUTF8(cls, data):  # Ensure proper encoding for UA's servers...
        """ Convert all strings to UTF-8 """
        for key in data:
            if isinstance(data[key], str):
                data[key] = data[key].encode('utf-8')
        return data

    # Apply stored properties to the given dataset & POST to the configured endpoint
    def send_data(self, session, values):
        logger.debug('get')
        logger.debug(values)

        return session.get(self.endpoint + '?' + urlencode(self.fixUTF8(values)),
                           headers={
                               'User-Agent': self.user_agent
                           })

    # Apply stored properties to the given dataset & POST to the configured endpoint
    async def send(self, values):
        logging.debug('send')
        logging.debug(values)

        logging.debug('self.user_agent')
        logging.debug(self.user_agent)

        loop = asyncio.get_event_loop()
        async with aiohttp.ClientSession(loop=loop) as session:
            async with self.send_data(self._session or session, values) as resp:
                logging.debug('status')
                logging.debug(resp.status)
                logging.debug('resp.text()')
                logging.debug(await resp.text())


class HTTPPost(HTTPRequest):
    def send_data(self, session, values):
        data = urllib.parse.urlencode(values)
        binary_data = data.encode('utf-8')

        logging.debug('binary_data')
        logging.debug(binary_data)

        return session.post(self.endpoint,
                            data=binary_data,
                            headers={
                                'User-Agent': self.user_agent
                            })


class Tracker:
    """ Primary tracking interface for Universal Analytics """
    params = None
    parameter_alias = {}
    valid_hittypes = ('pageview', 'event', 'social', 'screenview', 'transaction', 'item', 'exception', 'timing')

    @classmethod
    def alias(cls, typemap, base, *names):
        """ Declare an alternate (humane) name for a measurement protocol parameter """
        cls.parameter_alias[base] = (typemap, base)
        for i in names:
            cls.parameter_alias[i] = (typemap, base)

    @classmethod
    def coerceParameter(cls, name, value=None):
        if isinstance(name, str) and name[0] == '&':
            return name[1:], str(value)
        elif name in cls.parameter_alias:
            typecast, param_name = cls.parameter_alias.get(name)
            return param_name, typecast(value)
        else:
            raise KeyError('Parameter "{0}" is not recognized'.format(name))

    def payload(self, data):
        for key, value in data.items():
            try:
                yield self.coerceParameter(key, value)
            except KeyError:
                continue

    option_sequence = {
        'pageview': [(str, 'dp')],
        'event': [(str, 'ec'), (str, 'ea'), (str, 'el'), (int, 'ev')],
        'social': [(str, 'sn'), (str, 'sa'), (str, 'st')],
        'timing': [(str, 'utc'), (str, 'utv'), (str, 'utt'), (str, 'utl')]
    }

    @classmethod
    def consume_options(cls, data, hittype, args):
        """ Interpret sequential arguments related to known hittypes based on declared structures """
        opt_position = 0
        data['t'] = hittype  # integrate hit type parameter
        if hittype in cls.option_sequence:
            for expected_type, optname in cls.option_sequence[hittype]:
                if opt_position < len(args) and isinstance(args[opt_position], expected_type):
                    data[optname] = args[opt_position]
                opt_position += 1

    @classmethod
    def hittime(cls, timestamp=None, age=None, milliseconds=None):
        """ Returns an integer represeting the milliseconds offset for a given hit (relative to now) """
        if isinstance(timestamp, (int, float)):
            return int(Time.milliseconds_offset(Time.from_unix(timestamp, milliseconds=milliseconds)))
        if isinstance(timestamp, datetime.datetime):
            return int(Time.milliseconds_offset(timestamp))
        if isinstance(age, (int, float)):
            return int(age * 1000) + (milliseconds or 0)

    @property
    def account(self):
        return self.params.get('tid', None)

    def __init__(self, account, name=None, client_id=None, hash_client_id=False, user_id=None, user_agent=None,
                 use_post=True):
        # for debug purpose
        self._session = None
        if use_post is False:
            self.http = HTTPRequest(user_agent=user_agent)
        else:
            self.http = HTTPPost(user_agent=user_agent)

        self.params = {'v': 1, 'tid': account}

        if client_id is None:
            client_id = generate_uuid()

        self.params['cid'] = client_id

        self.hash_client_id = hash_client_id

        if user_id is not None:
            self.params['uid'] = user_id

    def set_timestamp(self, data):
        """ Interpret time-related options, apply queue-time parameter as needed """
        if 'hittime' in data:  # an absolute timestamp
            data['qt'] = self.hittime(timestamp=data.pop('hittime', None))
        if 'hitage' in data:  # a relative age (in seconds)
            data['qt'] = self.hittime(age=data.pop('hitage', None))

    async def send(self, hittype, *args, **data):
        """ Transmit HTTP requests to Google Analytics using the measurement protocol """

        if hittype not in self.valid_hittypes:
            raise KeyError('Unsupported Universal Analytics Hit Type: {0}'.format(repr(hittype)))

        self.set_timestamp(data)
        self.consume_options(data, hittype, args)

        for item in args:  # process dictionary-object arguments of transcient data
            if isinstance(item, dict):
                for key, val in self.payload(item):
                    data[key] = val

        for k, v in self.params.items():  # update only absent parameters
            if k not in data:
                data[k] = v

        data = dict(self.payload(data))

        if self.hash_client_id:
            data['cid'] = generate_uuid(data['cid'])

        # Transmit the hit to Google...
        await self.http.send(data)

    # Setting persistent attibutes of the session/hit/etc (inc. custom dimensions/metrics)
    def set(self, name, value=None):
        if isinstance(name, dict):
            for key, value in name.items():
                try:
                    param, value = self.coerceParameter(key, value)
                    self.params[param] = value
                except KeyError:
                    pass
        elif isinstance(name, str):
            try:
                param, value = self.coerceParameter(name, value)
                self.params[param] = value
            except KeyError:
                pass

    def __getitem__(self, name):
        param, value = self.coerceParameter(name, None)
        return self.params.get(param, None)

    def __setitem__(self, name, value):
        param, value = self.coerceParameter(name, value)
        self.params[param] = value

    def __delitem__(self, name):
        param, value = self.coerceParameter(name, None)
        if param in self.params:
            del self.params[param]


def safe_unicode(obj):
    """ Safe convertion to the Unicode string version of the object """
    try:
        return str(obj)
    except UnicodeDecodeError:
        return obj.decode('utf-8')


# Declaring name mappings for Measurement Protocol parameters
MAX_CUSTOM_DEFINITIONS = 200
MAX_EC_LISTS = 11  # 1-based index
MAX_EC_PRODUCTS = 11  # 1-based index
MAX_EC_PROMOTIONS = 11  # 1-based index

Tracker.alias(int, 'v', 'protocol-version')
Tracker.alias(safe_unicode, 'cid', 'client-id', 'clientId', 'clientid')
Tracker.alias(safe_unicode, 'tid', 'trackingId', 'account')
Tracker.alias(safe_unicode, 'uid', 'user-id', 'userId', 'userid')
Tracker.alias(safe_unicode, 'uip', 'user-ip', 'userIp', 'ipaddr')
Tracker.alias(safe_unicode, 'ua', 'userAgent', 'userAgentOverride', 'user-agent')
Tracker.alias(safe_unicode, 'dp', 'page', 'path')
Tracker.alias(safe_unicode, 'dt', 'title', 'pagetitle', 'pageTitle' 'page-title')
Tracker.alias(safe_unicode, 'dl', 'location')
Tracker.alias(safe_unicode, 'dh', 'hostname')
Tracker.alias(safe_unicode, 'sc', 'sessioncontrol', 'session-control', 'sessionControl')
Tracker.alias(safe_unicode, 'dr', 'referrer', 'referer')
Tracker.alias(int, 'qt', 'queueTime', 'queue-time')
Tracker.alias(safe_unicode, 't', 'hitType', 'hittype')
Tracker.alias(int, 'aip', 'anonymizeIp', 'anonIp', 'anonymize-ip')

# Campaign attribution
Tracker.alias(safe_unicode, 'cn', 'campaign', 'campaignName', 'campaign-name')
Tracker.alias(safe_unicode, 'cs', 'source', 'campaignSource', 'campaign-source')
Tracker.alias(safe_unicode, 'cm', 'medium', 'campaignMedium', 'campaign-medium')
Tracker.alias(safe_unicode, 'ck', 'keyword', 'campaignKeyword', 'campaign-keyword')
Tracker.alias(safe_unicode, 'cc', 'content', 'campaignContent', 'campaign-content')
Tracker.alias(safe_unicode, 'ci', 'campaignId', 'campaignID', 'campaign-id')

# Technical specs
Tracker.alias(safe_unicode, 'sr', 'screenResolution', 'screen-resolution', 'resolution')
Tracker.alias(safe_unicode, 'vp', 'viewport', 'viewportSize', 'viewport-size')
Tracker.alias(safe_unicode, 'de', 'encoding', 'documentEncoding', 'document-encoding')
Tracker.alias(int, 'sd', 'colors', 'screenColors', 'screen-colors')
Tracker.alias(safe_unicode, 'ul', 'language', 'user-language', 'userLanguage')

# Mobile app
Tracker.alias(safe_unicode, 'an', 'appName', 'app-name', 'app')
Tracker.alias(safe_unicode, 'cd', 'contentDescription', 'screenName', 'screen-name', 'content-description')
Tracker.alias(safe_unicode, 'av', 'appVersion', 'app-version', 'version')
Tracker.alias(safe_unicode, 'aid', 'appID', 'appId', 'application-id', 'app-id', 'applicationId')
Tracker.alias(safe_unicode, 'aiid', 'appInstallerId', 'app-installer-id')

# Ecommerce
Tracker.alias(safe_unicode, 'ta', 'affiliation', 'transactionAffiliation', 'transaction-affiliation')
Tracker.alias(safe_unicode, 'ti', 'transaction', 'transactionId', 'transaction-id')
Tracker.alias(float, 'tr', 'revenue', 'transactionRevenue', 'transaction-revenue')
Tracker.alias(float, 'ts', 'shipping', 'transactionShipping', 'transaction-shipping')
Tracker.alias(float, 'tt', 'tax', 'transactionTax', 'transaction-tax')
Tracker.alias(safe_unicode, 'cu', 'currency', 'transactionCurrency',
              'transaction-currency')  # Currency code, e.g. USD, EUR
Tracker.alias(safe_unicode, 'in', 'item-name', 'itemName')
Tracker.alias(float, 'ip', 'item-price', 'itemPrice')
Tracker.alias(float, 'iq', 'item-quantity', 'itemQuantity')
Tracker.alias(safe_unicode, 'ic', 'item-code', 'sku', 'itemCode')
Tracker.alias(safe_unicode, 'iv', 'item-variation', 'item-category', 'itemCategory', 'itemVariation')

# Events
Tracker.alias(safe_unicode, 'ec', 'event-category', 'eventCategory', 'category')
Tracker.alias(safe_unicode, 'ea', 'event-action', 'eventAction', 'action')
Tracker.alias(safe_unicode, 'el', 'event-label', 'eventLabel', 'label')
Tracker.alias(int, 'ev', 'event-value', 'eventValue', 'value')
Tracker.alias(int, 'ni', 'noninteractive', 'nonInteractive', 'noninteraction', 'nonInteraction')

# Social
Tracker.alias(safe_unicode, 'sa', 'social-action', 'socialAction')
Tracker.alias(safe_unicode, 'sn', 'social-network', 'socialNetwork')
Tracker.alias(safe_unicode, 'st', 'social-target', 'socialTarget')

# Exceptions
Tracker.alias(safe_unicode, 'exd', 'exception-description', 'exceptionDescription', 'exDescription')
Tracker.alias(int, 'exf', 'exception-fatal', 'exceptionFatal', 'exFatal')

# User Timing
Tracker.alias(safe_unicode, 'utc', 'timingCategory', 'timing-category')
Tracker.alias(safe_unicode, 'utv', 'timingVariable', 'timing-variable')
Tracker.alias(int, 'utt', 'time', 'timingTime', 'timing-time')
Tracker.alias(safe_unicode, 'utl', 'timingLabel', 'timing-label')
Tracker.alias(float, 'dns', 'timingDNS', 'timing-dns')
Tracker.alias(float, 'pdt', 'timingPageLoad', 'timing-page-load')
Tracker.alias(float, 'rrt', 'timingRedirect', 'timing-redirect')
Tracker.alias(safe_unicode, 'tcp', 'timingTCPConnect', 'timing-tcp-connect')
Tracker.alias(safe_unicode, 'srt', 'timingServerResponse', 'timing-server-response')

# Custom dimensions and metrics
for i in range(0, 200):
    Tracker.alias(safe_unicode, 'cd{0}'.format(i), 'dimension{0}'.format(i))
    Tracker.alias(int, 'cm{0}'.format(i), 'metric{0}'.format(i))

# Enhanced Ecommerce
Tracker.alias(str, 'pa')  # Product action
Tracker.alias(str, 'tcc')  # Coupon code
Tracker.alias(safe_unicode, 'pal')  # Product action list
Tracker.alias(int, 'cos')  # Checkout step
Tracker.alias(str, 'col')  # Checkout step option

Tracker.alias(str, 'promoa')  # Promotion action

for product_index in range(1, MAX_EC_PRODUCTS):
    Tracker.alias(str, 'pr{0}id'.format(product_index))  # Product SKU
    Tracker.alias(safe_unicode, 'pr{0}nm'.format(product_index))  # Product name
    Tracker.alias(safe_unicode, 'pr{0}br'.format(product_index))  # Product brand
    Tracker.alias(safe_unicode, 'pr{0}ca'.format(product_index))  # Product category
    Tracker.alias(safe_unicode, 'pr{0}va'.format(product_index))  # Product variant
    Tracker.alias(str, 'pr{0}pr'.format(product_index))  # Product price
    Tracker.alias(int, 'pr{0}qt'.format(product_index))  # Product quantity
    Tracker.alias(str, 'pr{0}cc'.format(product_index))  # Product coupon code
    Tracker.alias(int, 'pr{0}ps'.format(product_index))  # Product position

    for custom_index in range(MAX_CUSTOM_DEFINITIONS):
        Tracker.alias(str, 'pr{0}cd{1}'.format(product_index, custom_index))  # Product custom dimension
        Tracker.alias(int, 'pr{0}cm{1}'.format(product_index, custom_index))  # Product custom metric

    for list_index in range(1, MAX_EC_LISTS):
        Tracker.alias(str, 'il{0}pi{1}id'.format(list_index, product_index))  # Product impression SKU
        Tracker.alias(safe_unicode, 'il{0}pi{1}nm'.format(list_index, product_index))  # Product impression name
        Tracker.alias(safe_unicode, 'il{0}pi{1}br'.format(list_index, product_index))  # Product impression brand
        Tracker.alias(safe_unicode, 'il{0}pi{1}ca'.format(list_index, product_index))  # Product impression category
        Tracker.alias(safe_unicode, 'il{0}pi{1}va'.format(list_index, product_index))  # Product impression variant
        Tracker.alias(int, 'il{0}pi{1}ps'.format(list_index, product_index))  # Product impression position
        Tracker.alias(int, 'il{0}pi{1}pr'.format(list_index, product_index))  # Product impression price

        for custom_index in range(MAX_CUSTOM_DEFINITIONS):
            Tracker.alias(str, 'il{0}pi{1}cd{2}'.format(list_index, product_index,
                                                        custom_index))  # Product impression custom dimension
            Tracker.alias(int, 'il{0}pi{1}cm{2}'.format(list_index, product_index,
                                                        custom_index))  # Product impression custom metric

for list_index in range(1, MAX_EC_LISTS):
    Tracker.alias(safe_unicode, 'il{0}nm'.format(list_index))  # Product impression list name

for promotion_index in range(1, MAX_EC_PROMOTIONS):
    Tracker.alias(str, 'promo{0}id'.format(promotion_index))  # Promotion ID
    Tracker.alias(safe_unicode, 'promo{0}nm'.format(promotion_index))  # Promotion name
    Tracker.alias(str, 'promo{0}cr'.format(promotion_index))  # Promotion creative
    Tracker.alias(str, 'promo{0}ps'.format(promotion_index))  # Promotion position



import pytest

from . import tracker
from .. import mocktracker
from ... import di, Story

story = None


def teardown_function(function):
    story and story.clear()


def test_event():
    t = tracker.MockTracker()
    t.event()


def test_new_message():
    t = tracker.MockTracker()
    t.new_message()


def test_new_user():
    t = tracker.MockTracker()
    t.new_user()


def test_story():
    t = tracker.MockTracker()
    t.story()


def test_get_mock_tracker_as_dep():
    global story
    story = Story()
    story.use(mocktracker.MockTracker())

    with di.child_scope():
        @di.desc()
        class OneClass:
            @di.inject()
            def deps(self, tracker):
                self.tracker = tracker

        assert isinstance(di.injector.get('one_class').tracker, mocktracker.MockTracker)



from botstory import matchers, utils
from botstory.ast import story_context
import re


def get_text(ctx, default=None):
    return story_context.get_message_data(ctx, 'text', default=default)


def get_raw_text(ctx, default=None):
    return story_context.get_message_data(ctx, 'text', 'raw', default=default)


@matchers.matcher()
class Any:
    """
    filter any raw text
    """
    type = 'text.Any'

    def __init__(self):
        pass

    def validate(self, ctx):
        return get_raw_text(ctx)


@matchers.matcher()
class Equal:
    """
    filter equal raw text (case sensitive)
    """
    type = 'text.Equal'

    def __init__(self, test_string):
        self.test_string = test_string

    def validate(self, ctx):
        return get_raw_text(ctx) == self.test_string

    def serialize(self):
        return self.test_string

    @staticmethod
    def can_handle(data):
        return utils.is_string(data)

    @staticmethod
    def handle(data):
        return Equal(data)


@matchers.matcher()
class EqualCaseIgnore:
    """
    filter equal raw text (case in-sensitive)
    """
    type = 'text.EqualCaseIgnore'

    def __init__(self, test_string):
        self.test_string = test_string.lower()

    def validate(self, ctx):
        return get_raw_text(ctx, '').lower() == self.test_string

    def serialize(self):
        return self.test_string


@matchers.matcher()
class Match:
    type = 'text.Match'

    def __init__(self, pattern, flags=0):
        self.matcher = re.compile(pattern, flags=flags)

    def validate(self, ctx):
        raw_txt = get_raw_text(ctx)
        if not raw_txt:
            return False

        matches = self.matcher.findall(raw_txt)
        if len(matches) == 0:
            return False

        story_context.set_message_data(ctx, 'text', 'matches', matches)
        return True

    def serialize(self):
        return {
            'pattern': self.matcher.pattern,
            'flags': self.matcher.flags,
        }



"""
inspired by
https://github.com/KeepSafe/aiohttp/blob/master/examples/fake_server.py

"""

import aiohttp
from aiohttp import web
from aiohttp.test_utils import unused_port

from aiohttp.resolver import DefaultResolver
import logging
import pathlib
import socket
import ssl

logger = logging.getLogger(__name__)


def http_method(method, path):
    def wrapper(func):
        func.__method__ = method
        func.__path__ = path
        return func

    return wrapper


def head(path):
    return http_method('HEAD', path)


def get(path):
    return http_method('GET', path)


def delete(path):
    return http_method('DELETE', path)


def options(path):
    return http_method('OPTIONS', path)


def patch(path):
    return http_method('PATCH', path)


def post(path):
    return http_method('POST', path)


def put(path):
    return http_method('PUT', path)


def trace(path):
    return http_method('TRACE', path)


class FakeResolver:
    _LOCAL_HOST = {0: '127.0.0.1',
                   socket.AF_INET: '127.0.0.1',
                   socket.AF_INET6: '::1'}

    def __init__(self, fakes, *, loop):
        """fakes -- dns -> port dict"""
        self._fakes = fakes
        self._resolver = DefaultResolver(loop=loop)

    async def resolve(self, host, port=0, family=socket.AF_INET):
        fake_port = self._fakes.get(host)
        if fake_port is not None:
            return [{'hostname': host,
                     'host': self._LOCAL_HOST[family], 'port': fake_port,
                     'family': family, 'proto': 0,
                     'flags': socket.AI_NUMERICHOST}]
        else:
            return await self._resolver.resolve(host, port, family)


class FakeServer:
    def __init__(self, loop):
        self.loop = loop
        self.port = None
        self.history = []

        self.app = web.Application(loop=loop)
        for name in dir(self.__class__):
            func = getattr(self.__class__, name)
            if hasattr(func, '__method__'):
                self.app.router.add_route(func.__method__,
                                          func.__path__,
                                          getattr(self, name))

        self.app.middlewares.append(self.middleware_factory)
        self.handler = None
        self.server = None
        here = pathlib.Path(__file__)
        ssl_cert = here.parent / 'server.crt'
        ssl_key = here.parent / 'server.key'
        self.ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        self.ssl_context.load_cert_chain(str(ssl_cert), str(ssl_key))

    async def middleware_factory(self, app, handler):
        async def middleware(request):
            response = await handler(request)
            logger.debug('request: {}'.format(request))
            logger.debug('response: {}'.format(response))

            self.history.append({
                'request': request,
                'response': response,
            })
            return response

        return middleware

    async def start(self):
        port = unused_port()
        self.handler = self.app.make_handler()
        self.server = await self.loop.create_server(self.handler,
                                                    '127.0.0.1', port,
                                                    ssl=self.ssl_context)

        self.port = port
        return {self.ROOT_URI: port}

    async def stop(self):
        self.server.close()
        await self.server.wait_closed()
        await self.app.shutdown()
        await self.handler.finish_connections()
        await self.app.cleanup()

    async def __aenter__(self):
        info = await self.start()
        resolver = FakeResolver(info, loop=self.loop)
        self.connector = aiohttp.TCPConnector(loop=self.loop,
                                              resolver=resolver,
                                              verify_ssl=False)
        return self

    def session(self):
        return aiohttp.ClientSession(connector=self.connector, loop=self.loop)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()



from botstory.middlewares import option
import logging
import pytest
from . import option
from ... import matchers
from ...utils import answer, SimpleTrigger

logger = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_should_ask_with_options():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('How are you?')
        def one_story():
            @story.part()
            async def ask(ctx):
                return await story.ask(
                    'I feel fine. How about you?',
                    options=[{
                        'title': 'Good!',
                        'payload': {'health': 1},
                    }, {
                        'title': 'Ugly!',
                        'payload': {'health': -1},
                    }, {
                        'title': 'As usual',
                        'payload': {'health': 0},
                    }],
                    user=ctx['user'],
                )

            @story.part()
            def get_health(ctx):
                trigger.receive(option.get_option(ctx))

        await talk.pure_text('How are you?')
        await talk.option({'health': 1})
        assert trigger.result() == {'health': 1}


@pytest.mark.asyncio
async def test_validate_option():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(receive=option.Any())
        def one_story():
            @story.part()
            def store_option(ctx):
                trigger.passed()

        await talk.option({'engine': 'start'})
        assert trigger.is_triggered


@pytest.mark.asyncio
async def test_discard_not_valid_text_message():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(receive=option.Any())
        def one_story():
            @story.part()
            def store_option(ctx):
                trigger.passed()

        await talk.pure_text('Start engine!')
        assert not trigger.is_triggered


def test_serialize_option_match():
    m_old = option.Match('yellow')
    m_new = matchers.deserialize(matchers.serialize(m_old))
    assert isinstance(m_new, option.Match)
    assert m_new.option == m_old.option


@pytest.mark.asyncio
async def test_validate_only_option():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(receive=option.Match('green'))
        def one_story():
            @story.part()
            def store_option(ctx):
                trigger.passed()

        await talk.option('green')
        assert trigger.is_triggered


def test_serialize_on_start_option():
    m_old = option.OnStart()
    m_new = matchers.deserialize(matchers.serialize(m_old))
    assert isinstance(m_new, option.OnStart)



import pytest
from .. import mockdb
from ... import di, Story


def test_get_mockdb_as_dep():
    story = Story()

    story.use(mockdb.MockDB())

    with di.child_scope():
        @di.desc()
        class OneClass:
            @di.inject()
            def deps(self, storage):
                self.storage = storage

        assert isinstance(di.injector.get('one_class').storage, mockdb.MockDB)



"""
inspired by
https://github.com/KeepSafe/aiohttp/blob/master/examples/fake_server.py

"""

from aiohttp import web
import logging

from .server import FakeServer, get, post, delete

logger = logging.getLogger(__name__)

URI = 'https://graph.facebook.com{}'


class FakeFacebook(FakeServer):
    ROOT_URI = 'graph.facebook.com'

    @get('/v2.7/me')
    async def on_me(self, request):
        return web.json_response({
            "name": "John Doe",
            "id": "12345678901234567"
        })

    @get('/v2.7/me/friends')
    async def on_my_friends(self, request):
        return web.json_response({
            "data": [
                {
                    "name": "Bill Doe",
                    "id": "233242342342"
                },
                {
                    "name": "Mary Doe",
                    "id": "2342342343222"
                },
                {
                    "name": "Alex Smith",
                    "id": "234234234344"
                },
            ],
            "paging": {
                "cursors": {
                    "before": "QVFIUjRtc2c5NEl0ajN",
                    "after": "QVFIUlpFQWM0TmVuaDRad0dt",
                },
                "next": ("https://graph.facebook.com/v2.7/12345678901234567/"
                         "friends?access_token=EAACEdEose0cB")
            },
            "summary": {
                "total_count": 3
            }})

    @post('/v2.6/me/messages/')
    async def on_messages(self, request):
        return web.json_response({
            'status': 'ok',
        })

    @post('/v2.6/me/thread_settings')
    async def on_thread_settings_post(self, request):
        return web.json_response({
            'status': 'ok',
        })

    @delete('/v2.6/me/thread_settings')
    async def on_remove_thread_setting(self, request):
        return web.json_response({
            'status': 'ok',
        })



import logging
from .. import di
from ..integrations.mocktracker import tracker as tracker_module

logger = logging.getLogger(__name__)


@di.desc(reg=False)
class Users:
    def __init__(self):
        self._tracker = tracker_module.MockTracker()

    @di.inject()
    def add_tracker(self, tracker):
        logger.debug('add_tracker')
        logger.debug(tracker)
        self._tracker = tracker

    def on_new_user_comes(self, user):
        """
        each interface that create new user should call this method.
        as well if someone need to catch new user it is good place
        to do it

        :param user:
        :return:
        """
        self._tracker.new_user(user)

    def clear(self):
        self._tracker = tracker_module.MockTracker()



from . import desciption as desc_module, inject as inject_module, injector_service

__all__ = []

injector = injector_service.Injector()

bind = injector.bind
child_scope = injector.child_scope
clear_instances = injector.clear_instances
desc = desc_module.desc
get = injector.get
inject = inject_module.inject

__all__.extend([bind, child_scope, clear_instances, desc, inject])



import asyncio
import logging
from motor import motor_asyncio
from ... import di

logger = logging.getLogger(__name__)


@di.desc('storage', reg=False)
class MongodbInterface:
    """
    https://github.com/mongodb/motor
    """

    def __init__(self,
                 uri='localhost',
                 db_name='bots',
                 user_collection_name='user',
                 session_collection_name='session',
                 ):
        self.cx = None
        self.db = None
        self.session_collection = None
        self.user_collection = None
        self.uri = uri
        self.db_name = db_name
        self.session_collection_name = session_collection_name
        self.user_collection_name = user_collection_name

    async def start(self):
        loop = asyncio.get_event_loop()
        logger.debug('start')
        self.cx = motor_asyncio.AsyncIOMotorClient(self.uri, io_loop=loop)
        logger.debug(' create client for: {}'.format(self.uri))
        self.db = self.cx.get_database(self.db_name)
        logger.debug(' get db: {}'.format(self.db_name))
        self.session_collection = self.db.get_collection(self.session_collection_name)
        logger.debug(' get session collection: {}'.format(self.session_collection_name))
        self.user_collection = self.db.get_collection(self.user_collection_name)
        logger.debug(' get user collection: {}'.format(self.user_collection_name))

    async def stop(self):
        self.cx = None
        self.db = None
        self.session_collection = None
        self.user_collection = None

    async def clear_collections(self):
        await self.session_collection.drop()
        self.session_collection = self.db.get_collection(self.session_collection_name)
        await self.user_collection.drop()
        self.user_collection = self.db.get_collection(self.user_collection_name)

    async def get_session(self, **kwargs):
        session = await self.session_collection.find_one(kwargs)
        if not session:
            return None
        session.pop('lastModified', None)
        return session

    async def set_session(self, session):
        logger.info('set_session {}'.format(session))
        if '_id' in session:
            _id = session['_id']
            session.pop('lastModified', None)
            res = await self.session_collection.update_one(
                {'_id': _id},
                {
                    '$set': session,
                    '$currentDate': {
                        'lastModified': True
                    },
                },
            )

            if res.matched_count != 0:
                return _id

        return await self.session_collection.insert(session)

    async def new_session(self, user, **kwargs):
        logger.info('new_session for {}'.format(user))
        kwargs['user_id'] = kwargs.get('user_id', user['_id'])
        kwargs['stack'] = kwargs.get('stack', [])
        _id = await self.session_collection.insert(kwargs)
        return await self.session_collection.find_one({'_id': _id})

    async def get_user(self, **kwargs):
        if 'id' in kwargs:
            kwargs['_id'] = kwargs.get('id', None)
            del kwargs['id']
        user = await self.user_collection.find_one(kwargs)
        if not user:
            return None
        user.pop('lastModified', None)
        return user

    async def set_user(self, user):
        logger.info('set_user {}'.format(user))
        if '_id' in user:
            _id = user['_id']
            user.pop('lastModified', None)
            res = await self.user_collection.update_one(
                {'_id': _id},
                {
                    '$set': user,
                    '$currentDate': {
                        'lastModified': True
                    },
                },
            )
            if res.matched_count != 0:
                return _id
        return await self.user_collection.insert(user)

    # TODO: should be able to process dictionary
    async def new_user(self, **kwargs):
        logger.debug('store new user {}'.format(kwargs))
        _id = await self.user_collection.insert(kwargs)
        return await self.user_collection.find_one({'_id': _id})



import asyncio
import logging
import pytest
from unittest.mock import call

from . import di
from .integrations import mockdb, mockhttp
from .middlewares import location, text
from .utils import answer, SimpleTrigger

logger = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_should_run_sequence_of_parts():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('hi there!')
        def one_story():
            @story.part()
            def then(ctx):
                trigger_1.passed()

            @story.part()
            def then(ctx):
                trigger_2.passed()

        await talk.pure_text('hi there!')

        assert trigger_1.is_triggered
        assert trigger_2.is_triggered


@pytest.mark.asyncio
async def test_should_wait_for_answer_on_ask():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('hi there!')
        def one_story():
            @story.part()
            async def then(message):
                return await story.ask('How are you?', user=message['user'])

            @story.part()
            def then(message):
                trigger.passed()

        await say_pure_text('hi there!')

        assert not trigger.is_triggered

        await say_pure_text('Great!')

        assert trigger.is_triggered


@pytest.mark.asyncio
async def test_should_prevent_other_story_to_start_until_we_waiting_for_answer():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('hi there!')
        def first_story():
            @story.part()
            async def then_ask(ctx):
                return await story.ask('How are you?', user=ctx['user'])

            @story.part()
            async def then_trigger_2(ctx):
                trigger_2.passed()

        @story.on('Great!')
        def second_story():
            @story.part()
            def then_trigger_1(ctx):
                trigger_1.passed()

        await say_pure_text('hi there!')
        await say_pure_text('Great!')

        assert trigger_2.is_triggered
        assert not trigger_1.is_triggered


@pytest.mark.asyncio
async def test_should_start_next_story_after_current_finished():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('hi there!')
        def first_story():
            @story.part()
            async def then(ctx):
                return await story.ask('How are you?', user=ctx['user'])

            @story.part()
            def then(ctx):
                pass

        @story.on('Great!')
        def second_story():
            @story.part()
            def then(ctx):
                trigger.passed()

        await talk.pure_text('hi there!')
        await talk.pure_text('Great!')
        await talk.pure_text('Great!')

        assert trigger.is_triggered


@pytest.mark.asyncio
async def test_should_match_group_of_matchers_between_parts_of_story():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()

    with answer.Talk() as talk:
        say_location = talk(answer.location)
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('hi there!')
        def one_story():
            @story.part()
            def then(ctx):
                return [text.Any(), location.Any()]

            @story.part()
            def then(ctx):
                trigger_1.passed()
                return [text.Any(), location.Any()]

            @story.part()
            def then(ctx):
                trigger_2.passed()

        await say_pure_text('hi there!')
        await say_location({'lat': 1, 'lng': 1})
        await say_pure_text('hi there!')

        assert trigger_1.is_triggered
        assert trigger_2.is_triggered


@pytest.mark.asyncio
async def test_should_match_group_of_matchers_on_story_start():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(receive=['hi there!', location.Any()])
        def one_story():
            @story.part()
            def then(ctx):
                trigger.passed()

        await talk.pure_text('hi there!')
        await talk.location({'lat': 1, 'lng': 1})

        assert trigger.triggered_times == 2


@pytest.mark.asyncio
async def test_can_combine_async_with_sync_parts():
    async_trigger = SimpleTrigger()
    sync_trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('yo!')
        def one_story():
            @story.part()
            async def async_part(ctx):
                await asyncio.sleep(1)
                async_trigger.passed()

            @story.part()
            def sync_part(ctx):
                sync_trigger.passed()

        await talk.pure_text('yo!')
        assert async_trigger.is_triggered
        assert sync_trigger.is_triggered


@pytest.mark.asyncio
async def test_should_drop_session_once_found_that_it_does_not_match_current_story_structure():
    trigger = SimpleTrigger()
    another_trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('one')
        def one_story():
            @story.part()
            def async_part(ctx):
                return text.Any()

            @story.part()
            def sync_part(ctx):
                trigger.passed()

        @story.on('two')
        def another_story():
            @story.part()
            def sync_part(ctx):
                another_trigger.passed()

        await talk.pure_text('one')
        # now session differs from story struct
        talk.session['stack'][-1]['topic'] += '_'
        await talk.pure_text('two')
        assert not trigger.is_triggered
        assert another_trigger.is_triggered


@pytest.mark.asyncio
async def test_should_start_middlewares():
    with answer.Talk() as talk:
        story = talk.story
        story.use(mockdb.MockDB())
        http = story.use(mockhttp.MockHttpInterface())
        await story.start()
        http.start.assert_called_once_with()


@pytest.mark.asyncio
async def test_setup_should_config_facebook_options():
    with answer.Talk() as talk:
        story = talk.story
        db = story.use(mockdb.MockDB())
        http = story.use(mockhttp.MockHttpInterface())

        await story.setup()

        db.setup.assert_called_once_with()
        http.setup.assert_called_once_with()


@pytest.mark.asyncio
async def test_should_run_before_start_start_and_then_after_start(mocker):
    with di.child_scope():
        with answer.Talk() as talk:
            story = talk.story

            @di.desc()
            class MockExtension:
                handler = mocker.stub()

                async def before_start(self):
                    self.handler('before_start')

                async def start(self):
                    self.handler('start')

                async def after_start(self):
                    self.handler('after_start')

            ext = story.use(MockExtension())
            await story.start()
            ext.handler.assert_has_calls([
                call('before_start'),
                call('start'),
                call('after_start'),
            ])



from botstory import matchers
from botstory.ast import story_context
import logging

logger = logging.getLogger(__name__)


def get_sticker(ctx):
    return story_context.get_message_data(ctx, 'sticker_id')


ANY_LIKE = None
SMALL_LIKE = '369239263222822'
MEDIUM_LIKE = '369239343222814'
BIG_LIKE = '369239383222810'

LIKE_STICKERS = (SMALL_LIKE, MEDIUM_LIKE, BIG_LIKE)


@matchers.matcher()
class Any:
    type = 'Sticker.Any'

    def __init__(self):
        pass

    def validate(self, ctx):
        return get_sticker(ctx)


@matchers.matcher()
class Like:
    type = 'Sticker.Like'

    def __init__(self, size=ANY_LIKE):
        if size == ANY_LIKE:
            self.valid_likes = LIKE_STICKERS
        else:
            self.valid_likes = size

    def validate(self, ctx):
        return get_sticker(ctx) in self.valid_likes

    def serialize(self):
        return self.option

    @staticmethod
    def deserialize(option):
        return Match(option)



from botstory import Story
from botstory.ast import callable, forking, story_context
from botstory.middlewares import location, text
import pytest


@pytest.fixture
def build_mock_context():
    def factory(msg):
        story = Story()

        @story.on('Hi there!')
        def one_story():
            @story.part()
            async def start(ctx):
                await story.say('Where do you go?', user=ctx['user'])
                return forking.Switch({
                    'location': location.Any(),
                    'text': text.Any(),
                })

            @story.case(match='location')
            def location_case():
                @story.part()
                def store_location(ctx):
                    return text.Equal('hello!')

            @story.case(match='text')
            def text_case():
                @story.part()
                def store_location(ctx):
                    pass

            @story.part()
            def after_switch(ctx):
                return callable.EndOfStory('the end')

        return story_context.StoryContext(msg, story.stories_library)

    return factory


@pytest.mark.asyncio
async def test_execute_immutability(build_mock_context):
    ctx_before = build_mock_context({
        'session': {
            'stack': [{
                'data': None,
                'step': 0,
                'topic': 'one_story',
            }, {
                'data': None,
                'step': 0,
                'topic': 'location_case',
            }],
        },
        'user': None,
        'data': None,
    })

    ctx_after = await story_context.reducers.execute(ctx_before)
    assert ctx_after is not ctx_before
    assert ctx_after.waiting_for is not ctx_before.waiting_for
    assert ctx_after.message is not ctx_before.message
    assert ctx_after.message['session'] is not ctx_before.message['session']
    assert ctx_after.message['session']['stack'] is not ctx_before.message['session']['stack']
    assert ctx_after.message['session']['stack'][-1] is not ctx_before.message['session']['stack'][-1]


@pytest.mark.asyncio
async def test_execute_immutability_with_end_of_story(build_mock_context):
    ctx_before = build_mock_context({
        'session': {
            'data': None,
            'stack': [{
                'data': None,
                'step': 2,
                'topic': 'one_story',
            }],
        },
        'user': None,
    })

    ctx_after = await story_context.reducers.execute(ctx_before)
    assert ctx_after is not ctx_before
    assert ctx_after.waiting_for is not ctx_before.waiting_for
    assert ctx_after.message is not ctx_before.message
    assert ctx_after.get_user_data() is not ctx_before.get_user_data()


def test_iterate_storyline_immutability(build_mock_context):
    ctx_before = build_mock_context({
        'session': {
            'stack': [{
                'data': None,
                'step': 0,
                'topic': 'one_story',
            }],
        },
        'user': None,
        'data': None,
    })

    story_line = story_context.reducers.iterate_storyline(ctx_before)

    story_part = next(story_line)
    while True:
        try:
            assert story_part != ctx_before
            assert story_part.message is not ctx_before.message
            assert story_part.message['session'] is not ctx_before.message['session']
            assert story_part.message['session']['stack'] is not ctx_before.message['session']['stack']
            assert story_part.message['session']['stack'][-1] is not ctx_before.message['session']['stack'][-1]
            story_line.send(story_part)
        except StopIteration:
            break


def test_iterate_storyline_send_ctx_back(build_mock_context):
    ctx_before = build_mock_context({
        'session': {
            'stack': [{
                'data': None,
                'step': 0,
                'topic': 'one_story',
            }],
        },
        'user': None,
        'data': None,
    })

    story_line = story_context.reducers.iterate_storyline(ctx_before)
    ctx_after = next(story_line)
    ctx_before = ctx_after
    ctx_before.message['session']['stack'][-1]['data'] = 'hello world!'

    ctx_after = story_line.send(ctx_before)
    assert ctx_before.message['session']['stack'][-1]['data'] == \
           ctx_after.message['session']['stack'][-1]['data']
    compare_ctx(ctx_before, ctx_after)
    ctx_before = ctx_after

    ctx_after = story_line.send(ctx_before)
    compare_ctx(ctx_before, ctx_after)
    with pytest.raises(StopIteration):
        story_line.send(ctx_before)


def compare_ctx(ctx1, ctx2):
    assert ctx1 != ctx2
    assert ctx1.message != ctx2.message


def test_scope_out_immutability(build_mock_context):
    ctx_before = build_mock_context({
        'session': {
            'stack': [{
                'data': None,
                'step': 2,
                'topic': 'one_story',
            }],
        },
        'user': None,
        'data': None,
    })
    ctx_after = story_context.reducers.scope_out(ctx_before)
    assert ctx_after != ctx_before


def test_scope_in_immutability(build_mock_context):
    ctx_before = build_mock_context({
        'session': {
            'stack': [{
                'data': None,
                'step': 1,
                'topic': 'one_story',
            }, ],
        },
        'user': None,
        'data': None,
    })
    ctx_after = story_context.reducers.scope_in(ctx_before)
    assert ctx_after != ctx_before
    assert ctx_after.stack() != ctx_before.stack()



import asyncio
import logging

from .middlewares import any, location, text

logger = logging.getLogger(__name__)


class Chat:
    def __init__(self):
        self.interfaces = {}

    async def ask(self, body, quick_replies=None, options=None, user=None):
        """
        simple ask with predefined quick replies

        :param body:
        :param quick_replies: (optional) in form of
        {'title': <message>, 'payload': <any json>}
        :param options:
        :param user:
        :return:
        """
        await self.send_text_message_to_all_interfaces(
            recipient=user,
            text=body,
            quick_replies=quick_replies,
            options=options,
        )
        return any.Any()

    # TODO: move to middlewares/location/location.py and make async

    def ask_location(self, body, user):
        # TODO:
        # 1 ask user
        say(body, user)

        # 2 wait for answer
        # 3 process answer
        # 4 ask details once we not sure
        return [location.Any(), text.Any()]

    async def say(self, body, user, options):
        """
        say something to user

        :param body:
        :param user:
        :return:
        """
        return await self.send_text_message_to_all_interfaces(
            recipient=user, text=body, options=options)

    async def send_text_message_to_all_interfaces(self, *args, **kwargs):
        """
        TODO:
        we should know from where user has come and use right interface
        as well right interface can be chosen

        :param args:
        :param kwargs:
        :return:
        """
        logger.debug('async_send_text_message_to_all_interfaces')
        tasks = [interface.send_text_message(*args, **kwargs)
                 for _, interface in self.interfaces.items()]

        logger.debug('  tasks')
        logger.debug(tasks)

        res = [body for body in await asyncio.gather(*tasks)]
        logger.debug('  res')
        logger.debug(res)
        return res

    def add_interface(self, interface):
        logger.debug('add_interface')
        logger.debug(interface)
        self.interfaces[interface.type] = interface
        return interface

    def clear(self):
        self.interfaces = {}



from botstory.ast import story_context
from ... import matchers


def get_option(ctx):
    return story_context.get_message_data(ctx, 'option')


@matchers.matcher()
class Any:
    type = 'Option.Any'

    def __init__(self):
        pass

    def validate(self, ctx):
        return get_option(ctx)


@matchers.matcher()
class Match:
    type = 'Option.Match'

    def __init__(self, option):
        self.option = option

    def validate(self, ctx):
        return get_option(ctx) == self.option

    def serialize(self):
        return self.option

    @staticmethod
    def deserialize(option):
        return Match(option)


@matchers.matcher()
class OnStart:
    type = 'Option.OnStart'
    DEFAULT_OPTION_PAYLOAD = 'BOT_STORY.PUSH_GET_STARTED_BUTTON'

    def validate(self, ctx):
        return get_option(ctx) == self.DEFAULT_OPTION_PAYLOAD



import asyncio
from botstory.ast import story_context
import logging
from unittest import mock
import pytest

from . import messenger
from .. import commonhttp, mockdb, mockhttp
from ... import di, Story, utils
from ...middlewares import any, option, sticker

logger = logging.getLogger(__name__)

story = None


def teardown_function(function):
    logger.debug('tear down!')
    story.clear()


@pytest.mark.asyncio
async def test_send_text_message():
    user = utils.build_fake_user()

    global story
    story = Story()

    interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.start()

    await interface.send_text_message(
        recipient=user, text='hi!', quick_replies=None
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty1',
        },
        json={
            'message': {
                'text': 'hi!',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_truncate_long_message():
    user = utils.build_fake_user()

    global story
    story = Story()

    interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.start()

    very_long_message = 'very_long_message' * 100
    await interface.send_text_message(
        recipient=user,
        text=very_long_message,
        quick_replies=None,
        options={
            'overflow': 'cut'
        }
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty1',
        },
        json={
            'message': {
                'text': very_long_message[:640],
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_truncate_with_ellipsis_long_message_by_default():
    user = utils.build_fake_user()

    global story
    story = Story()

    interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.start()

    very_long_message = 'very_long_message' * 100
    await interface.send_text_message(
        recipient=user,
        text=very_long_message,
        quick_replies=None,
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty1',
        },
        json={
            'message': {
                'text': very_long_message[:638] + '\u2026',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_integration():
    user = utils.build_fake_user()

    global story
    story = Story()

    story.use(messenger.FBInterface(page_access_token='qwerty2'))
    story.use(mockdb.MockDB())
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.say('hi there!', user=user)

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty2',
        },
        json={
            'message': {
                'text': 'hi there!',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_quick_replies():
    user = utils.build_fake_user()

    global story
    story = Story()

    story.use(messenger.FBInterface(page_access_token='qwerty3'))
    story.use(mockdb.MockDB())
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.ask(
        'Which color do you like?',
        quick_replies=[{
            'title': 'Red',
            'payload': 0xff0000,
        }, {
            'title': 'Green',
            'payload': 0x00ff00,
        }, {
            'title': 'Blue',
            'payload': 0x0000ff,
        }],
        user=user,
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty3',
        },
        json={
            'message': {
                'text': 'Which color do you like?',
                'quick_replies': [
                    {
                        'content_type': 'text',
                        'title': 'Red',
                        'payload': 0xff0000,
                    },
                    {
                        'content_type': 'text',
                        'title': 'Green',
                        'payload': 0x00ff00,
                    },
                    {
                        'content_type': 'text',
                        'title': 'Blue',
                        'payload': 0x0000ff,
                    },
                ],
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_setup_webhook():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.start()

    mock_http.webhook.assert_called_with(
        '/webhook',
        fb_interface.handle,
        'some-token',
    )


@pytest.mark.asyncio
async def test_should_request_user_data_once_we_do_not_know_current_user():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(
        page_access_token='qwerty4',
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    http = story.use(mockhttp.MockHttpInterface(get={
        'first_name': 'Peter',
        'last_name': 'Chang',
        'profile_pic': 'https://fbcdn-profile-a.akamaihd.net/hprofile-ak-xpf1/v/t1.0-1/p200x200/13055603_10105219398495383_8237637584159975445_n.jpg?oh=1d241d4b6d4dac50eaf9bb73288ea192&oe=57AF5C03&__gda__=1470213755_ab17c8c8e3a0a447fed3f272fa2179ce',
        'locale': 'en_US',
        'timezone': -7,
        'gender': 'male'
    }))
    story.use(mockdb.MockDB())

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    http.get.assert_called_with(
        'https://graph.facebook.com/v2.6/USER_ID',
        params={
            'access_token': 'qwerty4',
        },
    )


@pytest.mark.asyncio
async def test_should_request_user_data_and_fail():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(
        page_access_token='qwerty5',
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    story.use(mockhttp.MockHttpInterface(
        get_raise=commonhttp.errors.HttpRequestError()))
    db = story.use(mockdb.MockDB())

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    assert (await db.get_user(facebook_user_id='USER_ID')).no_fb_profile is True


@pytest.mark.asyncio
async def test_webhook_handler_should_return_ok_status_if_http_fail():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(
        page_access_token='qwerty6',
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    story.use(mockhttp.MockHttpInterface(get_raise=commonhttp.errors.HttpRequestError()))
    story.use(mockdb.MockDB())

    res = await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    assert res['status'] == 200


@pytest.mark.asyncio
async def test_webhook_handler_should_return_ok_status_in_any_case():
    global story
    story = Story()

    fb_interface = messenger.FBInterface()
    with mock.patch('botstory.integrations.fb.messenger.logger') as mock_logger:
        res = await fb_interface.handle({
            'object': 'page',
            'entry': [{
                'id': 'PAGE_ID',
                'time': 1473204787206,
                'messaging': [
                    {
                        'sender': {
                            'id': 'USER_ID'
                        },
                        'recipient': {
                            'id': 'PAGE_ID'
                        },
                        'timestamp': 1458692752478,
                        'message': {
                            'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                            'seq': 73,
                            'text': 'hello, world!'
                        }
                    }
                ]
            }]
        })

        assert mock_logger.debug.calledWith()

    assert res['status'] == 200


# integration

@pytest.fixture
def build_fb_interface():
    async def builder():
        user = utils.build_fake_user()
        session = utils.build_fake_session()

        global story
        story = Story()

        storage = story.use(mockdb.MockDB())
        fb = story.use(messenger.FBInterface(page_access_token='qwerty'))

        await story.start()

        await storage.set_session(session)
        await storage.set_user(user)

        return fb, story

    return builder


@pytest.mark.asyncio
async def test_handler_raw_text(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on('hello, world!')
    def correct_story():
        @story.part()
        def store_result(ctx):
            correct_trigger.receive(story_context.get_message_data(ctx))

    @story.on('Goodbye, world!')
    def incorrect_story():
        @story.part()
        def store_result(ctx):
            incorrect_trigger.receive(story_context.get_message_data(ctx))

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'text': {
            'raw': 'hello, world!'
        }
    }


@pytest.mark.asyncio
async def test_handler_selected_option(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on(receive=option.Match('GREEN'))
    def correct_story():
        @story.part()
        def store_result(ctx):
            correct_trigger.receive(story_context.get_message_data(ctx))

    @story.on(receive=option.Match('BLUE'))
    def incorrect_story():
        @story.part()
        def store_result(ctx):
            incorrect_trigger.receive(story_context.get_message_data(ctx))

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [{
                'sender': {
                    'id': 'USER_ID'
                },
                'recipient': {
                    'id': 'PAGE_ID'
                },
                'timestamp': 1458692752478,
                'message': {
                    'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                    'seq': 73,
                    'text': 'Green!',
                    'quick_reply': {
                        'payload': 'GREEN'
                    }
                }
            }]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'option': 'GREEN',
        'text': {
            'raw': 'Green!'
        }
    }


@pytest.mark.asyncio
async def test_handler_postback(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on(receive=option.Match('GREEN'))
    def correct_story():
        @story.part()
        def store_result(ctx):
            correct_trigger.receive(story_context.get_message_data(ctx))

    @story.on(receive=option.Match('BLUE'))
    def incorrect_story():
        @story.part()
        def store_result(ctx):
            incorrect_trigger.receive(story_context.get_message_data(ctx))

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [{
                'sender': {
                    'id': 'USER_ID'
                },
                'recipient': {
                    'id': 'PAGE_ID'
                },
                'timestamp': 1458692752478,
                'postback': {
                    'payload': 'GREEN'
                },
            }]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'option': 'GREEN',
    }


@pytest.mark.asyncio
async def test_handler_thumbsup(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    like_is_here_trigger = utils.SimpleTrigger()

    @story.on(receive=sticker.Like())
    def like_story():
        @story.part()
        def store_result(ctx):
            like_is_here_trigger.passed()

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [{
                'sender': {
                    'id': 'USER_ID'
                },
                'recipient': {
                    'id': 'PAGE_ID'
                },
                'timestamp': 1458692752478,
                "message": {
                    'sticker_id': sticker.SMALL_LIKE,
                }
            }]
        }]
    })

    assert like_is_here_trigger.is_passed()


@pytest.mark.asyncio
async def test_should_not_process_echo_delivery_and_read_messages_as_regular(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    echo_trigger = utils.SimpleTrigger()

    @story.on(receive=any.Any())
    def one_story():
        @story.part()
        def sync_part(message):
            echo_trigger.passed()

    await fb_interface.handle({
        'entry': [
            {
                'id': '329188380752158',
                'messaging': [{
                    'message': {
                        'app_id': 345865645763384,
                        'is_echo': 'True',
                        'mid': 'mid.1477350590023:38b1efd593',
                        'seq': 323,
                        'text': 'Hm I dont know what is it'
                    },
                    'recipient': {
                        'id': '1034692249977067'
                    },
                    'sender': {
                        'id': '329188380752158'
                    },
                    'timestamp': 1477350590023
                }, {
                    'read': {
                        'seq': 2697,
                        'watermark': 1477354670744
                    },
                    'recipient': {
                        'id': '329188380752158'
                    },
                    'sender': {
                        'id': '1034692249977067'
                    },
                    'timestamp': 1477354672037
                }, {
                    'delivery': {
                        'mids': [
                            'mid.1477354667117:8fedc43d37'
                        ],
                        'seq': 2679,
                        'watermark': 1477354668538
                    },
                    'recipient': {
                        'id': '329188380752158'
                    },
                    'sender': {
                        'id': '1034692249977067'
                    },
                    'timestamp': 0
                }],
                'time': 1477350590772
            }
        ],
        'object': 'page'
    })

    assert not echo_trigger.is_triggered


@pytest.mark.asyncio
async def test_set_greeting_text():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty7'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.set_greeting_text('Hi there {{user_first_name}}!')

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty7',
        },
        json={
            'setting_type': 'greeting',
            'greeting': {
                'text': 'Hi there {{user_first_name}}!',
            },
        }
    )


@pytest.mark.asyncio
async def test_can_set_greeting_text_before_inject_http():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty8'))
    await fb_interface.set_greeting_text('Hi there {{user_first_name}}!')

    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.setup()

    # give few a moment for lazy initialization of greeting text
    await asyncio.sleep(0.1)

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty8',
        },
        json={
            'setting_type': 'greeting',
            'greeting': {
                'text': 'Hi there {{user_first_name}}!',
            },
        }
    )


@pytest.mark.asyncio
async def test_can_set_greeting_text_in_constructor():
    global story
    story = Story()

    fb = story.use(messenger.FBInterface(
        greeting_text='Hi there {{user_first_name}}!',
        page_access_token='qwerty9',
    ))

    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.setup()

    # give few a moment for lazy initialization of greeting text
    await asyncio.sleep(0.1)

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty9',
        },
        json={
            'setting_type': 'greeting',
        },
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty9',
        },
        json={
            'setting_type': 'greeting',
            'greeting': {
                'text': 'Hi there {{user_first_name}}!',
            },
        }
    )


@pytest.mark.asyncio
async def test_remove_greeting_text():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty10'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.remove_greeting_text()

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty10',
        },
        json={
            'setting_type': 'greeting',
        }
    )


@pytest.mark.asyncio
async def test_set_greeting_call_to_action_payload():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty11'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.set_greeting_call_to_action_payload('SOME_PAYLOAD')

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty11',
        },
        json={
            'setting_type': 'call_to_actions',
            'thread_state': 'new_thread',
            'call_to_actions': [{'payload': 'SOME_PAYLOAD'}]
        }
    )


@pytest.mark.asyncio
async def test_remove_greeting_call_to_action_payload():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty12'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.remove_greeting_call_to_action_payload()

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty12',
        },
        json={
            'setting_type': 'call_to_actions',
            'thread_state': 'new_thread',
        }
    )


@pytest.mark.asyncio
async def test_set_persistent_menu():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty13'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.set_persistent_menu([{
        'type': 'postback',
        'title': 'Help',
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
    }, {
        'type': 'web_url',
        'title': 'View Website',
        'url': 'http://petersapparel.parseapp.com/'
    }])

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty13',
        },
        json={
            'setting_type': 'call_to_actions',
            'thread_state': 'existing_thread',
            'call_to_actions': [{
                'type': 'postback',
                'title': 'Help',
                'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
            }, {
                'type': 'web_url',
                'title': 'View Website',
                'url': 'http://petersapparel.parseapp.com/'
            }]
        }
    )


@pytest.mark.asyncio
async def test_can_set_persistent_menu_before_http():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty14'))
    await fb_interface.set_persistent_menu([{
        'type': 'postback',
        'title': 'Help',
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
    }, {
        'type': 'web_url',
        'title': 'View Website',
        'url': 'http://petersapparel.parseapp.com/'
    }])

    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.setup()

    # give few a moment for lazy initialization of greeting text
    await asyncio.sleep(0.1)

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty14',
        },
        json={
            'setting_type': 'call_to_actions',
            'thread_state': 'existing_thread',
            'call_to_actions': [{
                'type': 'postback',
                'title': 'Help',
                'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
            }, {
                'type': 'web_url',
                'title': 'View Website',
                'url': 'http://petersapparel.parseapp.com/'
            }]
        }
    )


@pytest.mark.asyncio
async def test_can_set_persistent_menu_inside_of_constructor():
    global story
    story = Story()

    story.use(messenger.FBInterface(
        page_access_token='qwerty15',
        persistent_menu=[{
            'type': 'postback',
            'title': 'Help',
            'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
        }, {
            'type': 'web_url',
            'title': 'View Website',
            'url': 'http://petersapparel.parseapp.com/'
        }]
    ))

    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.setup()

    # give few a moment for lazy initialization of greeting text
    await asyncio.sleep(0.1)

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty15',
        },
        json={
            'setting_type': 'call_to_actions',
            'thread_state': 'existing_thread',
        }
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty15',
        },
        json={
            'setting_type': 'call_to_actions',
            'thread_state': 'existing_thread',
            'call_to_actions': [{
                'type': 'postback',
                'title': 'Help',
                'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
            }, {
                'type': 'web_url',
                'title': 'View Website',
                'url': 'http://petersapparel.parseapp.com/'
            }]
        }
    )


@pytest.mark.asyncio
async def test_remove_persistent_menu():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty16'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.remove_persistent_menu()

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/thread_settings',
        params={
            'access_token': 'qwerty16',
        },
        json={
            'setting_type': 'call_to_actions',
            'thread_state': 'existing_thread'
        }
    )


def test_get_fb_as_deps():
    global story
    story = Story()

    story.use(messenger.FBInterface())

    with di.child_scope():
        @di.desc()
        class OneClass:
            @di.inject()
            def deps(self, fb):
                self.fb = fb

        assert isinstance(di.injector.get('one_class').fb, messenger.FBInterface)


def test_bind_fb_deps():
    global story
    story = Story()

    story.use(messenger.FBInterface())
    story.use(mockdb.MockDB())
    story.use(mockhttp.MockHttpInterface())

    with di.child_scope():
        @di.desc()
        class OneClass:
            @di.inject()
            def deps(self, fb):
                self.fb = fb

        assert isinstance(di.injector.get('one_class').fb.http, mockhttp.MockHttpInterface)
        assert isinstance(di.injector.get('one_class').fb.storage, mockdb.MockDB)



#!/usr/bin/python
# -*- encoding: utf-8 -*-
###############################################################################
# Test and example kit for Universal Analytics for Python
# Copyright (c) 2013, Analytics Pros
#
# This project is free software, distributed under the BSD license.
# Analytics Pros offers consulting and integration services if your firm needs
# assistance in strategy, implementation, or auditing existing work.
###############################################################################

import pytest
import urllib

from .tracker import Tracker
from . import fake_analytics


def url_quote(value, safe_chars=''):
    return urllib.parse.quote(value, safe_chars)


@pytest.fixture
def tracker():
    return Tracker('UA-XXXXX-Y', use_post=True)


def test_tracker_options_basic(tracker):
    assert 'UA-XXXXX-Y' == tracker.params['tid']


def test_persistent_campaign_settings(tracker):
    # Apply campaign settings
    tracker.set('campaignName', 'testing-campaign')
    tracker.set('campaignMedium', 'testing-medium')
    tracker['campaignSource'] = 'test-source'

    tracker.params['cn'] = 'testing-campaign'
    tracker.params['cm'] = 'testing-medium'
    tracker.params['cs'] = 'test-source'


@pytest.mark.asyncio
async def test_send_pageview(tracker, event_loop):
    async with fake_analytics.FakeAnalytics(event_loop) as server:
        async with server.session() as session:
            tracker.http._session = session

            await tracker.send('pageview', '/test')

            assert len(server.history) == 1
            text = await server.history[-1]['request'].text()
            assert 't=pageview' in text
            assert 'dp={0}'.format(url_quote('/test')) in text


@pytest.mark.asyncio
async def test_send_interactive_event(tracker, event_loop):
    async with fake_analytics.FakeAnalytics(event_loop) as server:
        async with server.session() as session:
            tracker.http._session = session

            await tracker.send('event', 'mycat', 'myact', 'mylbl', {'noninteraction': 1, 'page': '/1'})

            assert len(server.history) == 1
            text = await server.history[-1]['request'].text()
            assert 't=event' in text
            assert 'ec=mycat' in text
            assert 'ea=myact' in text
            assert 'el=mylbl' in text
            assert 'ni=1' in text
            assert 'dp={0}'.format(url_quote('/1')) in text


# def testSendUnicodeEvent(self):
#         # Send unicode data:
#         # As unicode
#         self.tracker.send('event', u'câtēgøry', u'åctîõn', u'låbęl', u'válüē')
#         # As str
#         self.tracker.send('event', 'câtēgøry', 'åctîõn', 'låbęl', 'válüē')
#
#         # TODO  write assertions for these...
#         # The output buffer should show both representations in UTF-8 for compatibility
#

@pytest.mark.asyncio
async def test_send_social_hit(tracker, event_loop):
    async with fake_analytics.FakeAnalytics(event_loop) as server:
        async with server.session() as session:
            tracker.http._session = session

            # Send a social hit
            await tracker.send('social', 'facebook', 'share', '/test#social')
            text = await server.history[-1]['request'].text()
            assert 't=social' in text
            assert 'sn=facebook' in text
            assert 'sa=share' in text
            assert 'st={0}'.format(url_quote('/test#social')) in text


@pytest.mark.asyncio
async def test_send_transaction(tracker, event_loop):
    async with fake_analytics.FakeAnalytics(event_loop) as server:
        async with server.session() as session:
            tracker.http._session = session

            # Dispatch the item hit first (though this is somewhat unusual)
            await tracker.send('item', {
                'transactionId': '12345abc',
                'itemName': 'pizza',
                'itemCode': 'abc',
                'itemCategory': 'hawaiian',
                'itemQ7uantity': 1
            }, hitage=200)

            # Then the transaction hit...
            await tracker.send('transaction', {
                'transactionId': '12345abc',
                'transactionAffiliation': 'phone order',
                'transactionRevenue': 28.00,
                'transactionTax': 3.00,
                'transactionShipping': 0.45,
                'transactionCurrency': 'USD'
            }, hitage=7200)

            # TODO: asserts


@pytest.mark.asyncio
async def test_send_transaction(tracker, event_loop):
    async with fake_analytics.FakeAnalytics(event_loop) as server:
        async with server.session() as session:
            tracker.http._session = session

            # A few more hits for good measure, testing real-time support for time offset
            await tracker.send('pageview', '/test', {'campaignName': 'testing2'}, hitage=60 * 5)  # 5 minutes ago
            await tracker.send('pageview', '/test', {'campaignName': 'testing3'}, hitage=60 * 20)  # 20 minutes ago

            # TODO: asserts



from botstory import matchers
from botstory.ast import callable, loop, stack_utils, story_context
import logging
import inspect

logger = logging.getLogger(__name__)


def enter_new_scope(ctx):
    """
    we inside new scope with it onw
    :param ctx:
    :return:
    """
    ctx = ctx.clone()
    ctx.waiting_for = ctx.compiled_story().children_matcher()
    return ctx


async def execute(ctx):
    """
    execute story part at the current context
    and make one step further

    :param ctx:
    :return:
    """
    tail_depth = len(ctx.stack()) - 1

    story_part = ctx.get_current_story_part()
    logger.debug('# going to call: {}'.format(story_part.__name__))
    waiting_for = story_part(ctx.message)
    logger.debug('# got result {}'.format(waiting_for))
    if inspect.iscoroutinefunction(story_part):
        waiting_for = await waiting_for

    # story part could run callable story and return its context
    if isinstance(waiting_for, story_context.StoryContext):
        # for such cases is very important to know `tail_depth`
        # because story context from callable story already has
        # few stack items above our tail
        ctx = waiting_for.clone()
        ctx.waiting_for = callable.WaitForReturn()
    else:
        ctx = ctx.clone()
        ctx.waiting_for = waiting_for

    tail_data = ctx.message['session']['stack'][tail_depth]['data']
    tail_step = ctx.message['session']['stack'][tail_depth]['step']
    if ctx.is_waiting_for_input():
        if isinstance(ctx.waiting_for, callable.EndOfStory):
            if isinstance(ctx.waiting_for.data, dict):
                new_data = {**ctx.get_user_data(), **ctx.waiting_for.data}
            else:
                new_data = ctx.waiting_for.data
            ctx.message = {
                **ctx.message,
                'session': {
                    **ctx.message['session'],
                    'data': new_data,
                },
            }
            tail_step += 1
        elif isinstance(ctx.waiting_for, loop.ScopeMatcher):
            # jumping in a loop
            tail_data = matchers.serialize(ctx.waiting_for)
        elif isinstance(ctx.waiting_for, loop.BreakLoop):
            tail_step += 1
        else:
            tail_data = matchers.serialize(
                matchers.get_validator(ctx.waiting_for)
            )
            tail_step += 1

    ctx.message = modify_stack_in_message(ctx.message,
                                          lambda stack: stack[:tail_depth] +
                                                        [{
                                                            'data': tail_data,
                                                            'step': tail_step,
                                                            'topic': stack[tail_depth]['topic'],
                                                        }] +
                                                        stack[tail_depth + 1:])
    logger.debug('# mutated ctx after execute')
    logger.debug(ctx)
    return ctx


def iterate_storyline(ctx):
    """
    iterate the last storyline from the last visited story part

    :param ctx:
    :return:
    """
    logger.debug('# start iterate')
    compiled_story = ctx.compiled_story()
    if not compiled_story:
        return

    for step in range(ctx.current_step(),
                      len(compiled_story.story_line)):
        ctx = ctx.clone()
        tail = ctx.stack_tail()
        ctx.message = modify_stack_in_message(ctx.message,
                                              lambda stack: stack[:-1] + [{
                                                  'data': tail['data'],
                                                  'step': step,
                                                  'topic': tail['topic'],
                                              }])

        logger.debug('# [{}] iterate'.format(step))
        logger.debug(ctx)

        ctx = yield ctx


def scope_in(ctx):
    """
    - build new scope on the top of stack
    - and current scope will wait for it result

    :param ctx:
    :return:
    """
    logger.debug('# scope_in')
    logger.debug(ctx)
    ctx = ctx.clone()

    compiled_story = None
    if not ctx.is_empty_stack():
        compiled_story = ctx.get_child_story()
        # we match child story loop once by message
        # what should prevent multiple matching by the same message
        ctx.matched = True
        ctx.message = modify_stack_in_message(ctx.message,
                                              lambda stack: stack[:-1] + [{
                                                  'data': matchers.serialize(callable.WaitForReturn()),
                                                  'step': stack[-1]['step'],
                                                  'topic': stack[-1]['topic']
                                              }])

    try:
        if not compiled_story and ctx.is_scope_level_part():
            compiled_story = ctx.get_current_story_part()
    except story_context.MissedStoryPart:
        pass

    if not compiled_story:
        compiled_story = ctx.compiled_story()

    logger.debug('# [>] going deeper')
    ctx.message = modify_stack_in_message(ctx.message,
                                          lambda stack: stack + [
                                              stack_utils.build_empty_stack_item(compiled_story.topic)])
    logger.debug(ctx)
    return ctx


def scope_out(ctx):
    """
    drop last stack item if:
     - we have reach the end of stack
     - and don't wait any input

    :param ctx:
    :return:
    """
    logger.debug('# scope_out')
    logger.debug(ctx)
    # we reach the end of story line
    # so we could collapse previous scope and related stack item
    if ctx.is_tail_of_story() and ctx.could_scope_out():
        logger.debug('# [<] return')
        ctx = ctx.clone()
        ctx.message['session']['stack'] = ctx.message['session']['stack'][:-1]
        if not ctx.is_empty_stack() and \
                (ctx.is_scope_level_part() or \
                         ctx.is_breaking_a_loop()):
            # isinstance(ctx.get_current_story_part(), loop.StoriesLoopNode) and \
            # isinstance(ctx.waiting_for, callable.EndOfStory) or \
            ctx.message = modify_stack_in_message(ctx.message,
                                                  lambda stack: stack[:-1] + [{
                                                      'data': stack[-1]['data'],
                                                      'step': stack[-1]['step'] + 1,
                                                      'topic': stack[-1]['topic'],
                                                  }])
            if ctx.is_breaking_a_loop() and not ctx.is_scope_level():
                ctx.waiting_for = None

        logger.debug(ctx)

    return ctx


def modify_stack_in_message(message, mutator):
    return {
        **message,
        'session': {
            **message['session'],
            'stack': mutator(message['session']['stack']),
        }
    }



class HttpRequestError(Exception):
    """HTTP error.
    Shortcut for raising HTTP errors with custom code, message and headers.
    :param int code: HTTP Error code.
    :param str message: (optional) Error message.
    :param list of [tuple] headers: (optional) Headers to be sent in response.

    based on https://github.com/KeepSafe/aiohttp/blob/v1.0.5/aiohttp/errors.py#L72
    """

    code = 0
    message = ''
    headers = None

    def __init__(self, *, code=None, message='', headers=None):
        if code is not None:
            self.code = code
        self.headers = headers
        self.message = message

        super().__init__("%s, message='%s'" % (self.code, message))



import inspect

from .parser import camel_case_to_underscore

from .. import di


def inject(t=None):
    def decorated_fn(fn):
        if inspect.isclass(fn):
            name = t or camel_case_to_underscore(fn.__name__)[0]
            print('register {} on name {}'.format(fn, name))
            di.injector.register(name, fn)
        elif inspect.isfunction(fn):
            di.injector.requires(fn)
        else:
            # I'm not sure whether it possible case
            raise NotImplementedError('try decorate {}'.format(fn))
        return fn

    return decorated_fn



import pytest
from .. import mockhttp
from ... import di, Story


def test_get_mockhttp_as_dep():
    story = Story()
    story.use(mockhttp.MockHttpInterface())

    with di.child_scope():
        @di.desc()
        class OneClass:
            @di.inject()
            def deps(self, http):
                self.http = http

        assert isinstance(di.injector.get('one_class').http, mockhttp.MockHttpInterface)



from botstory.ast import stack_utils, story_context
import logging
from .. import matchers

logger = logging.getLogger(__name__)


@matchers.matcher()
class WaitForReturn:
    type = 'WaitForReturn'

    def __init__(self):
        pass

    def validate(self, message):
        # return 'return' in message
        return True


class EndOfStory:
    type = 'EndOfStory'

    def __init__(self, data={}):
        self.data = data


class CallableNodeWrapper:
    """
    helps start processing callable story
    """

    def __init__(self, ast_node, library, processor):
        self.library = library
        self.ast_node = ast_node
        self.processor = processor

    async def startpoint(self, *args, **kwargs):
        if len(args) > 0:
            parent_ctx = args[0]
            session = parent_ctx['session']
            user = parent_ctx['user']
        else:
            if {'session', 'user'} > set(kwargs):
                raise AttributeError('Got {} and {}. Should pass session as well'.format(args, kwargs))
            session = kwargs.pop('session')
            user = kwargs.pop('user')

        # we are going deeper so prepare one more item in stack
        logger.debug('  action: extend stack by +1')
        session = {
            **session,
            'data': kwargs,
            'stack': session['stack'] + [stack_utils.build_empty_stack_item(self.ast_node.topic)],
        }

        ctx = story_context.StoryContext(message={
            'session': session,
            # TODO: should get user from context
            'user': user,
        }, library=self.library)
        ctx = await self.processor.process_story(ctx)
        ctx = story_context.reducers.scope_out(ctx)

        logger.debug('# result of callable')
        logger.debug(ctx)

        if isinstance(ctx.waiting_for, EndOfStory):
            return ctx.waiting_for.data

        # because we would like to return to stories from caller
        # we should return our context to callee
        return ctx


class CallableStoriesAPI:
    def __init__(self, library, parser_instance, processor_instance):
        self.library = library
        self.parser_instance = parser_instance
        self.processor_instance = processor_instance

    def callable(self):
        def fn(callable_story):
            compiled_story = self.parser_instance.compile(
                callable_story,
            )
            self.library.add_callable(compiled_story)
            return CallableNodeWrapper(
                compiled_story,
                self.library,
                self.processor_instance,
            ).startpoint

        return fn



import logging
import pytest
from botstory import matchers, utils
from botstory.middlewares import sticker
from botstory.utils import answer

logger = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_match_any_sticker():
    trigger = utils.SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(sticker.Any())
        def sticker_story():
            @story.part()
            def then(ctx):
                trigger.passed()

        await talk.ask({
            'sticker_id': sticker.LIKE_STICKERS[0],
        })

        assert trigger.is_triggered


@pytest.mark.asyncio
async def test_match_fblike_sticker():
    trigger_any = utils.SimpleTrigger()
    trigger_like = utils.SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(sticker.Like())
        def like_sticker_story():
            @story.part()
            def then(ctx):
                trigger_like.passed()

        @story.on(sticker.Any())
        def any_sticker_story():
            @story.part()
            def then(ctx):
                trigger_any.passed()

        await talk.ask({
            'sticker_id': sticker.LIKE_STICKERS[0],
        })

        assert trigger_like.is_triggered

        await talk.ask({
            'sticker_id': '1',
        })

        assert trigger_like.is_triggered


@pytest.mark.parametrize(('match_size', 'message_size'),
                         [(sticker.SMALL_LIKE, sticker.SMALL_LIKE),
                          (sticker.MEDIUM_LIKE, sticker.MEDIUM_LIKE),
                          (sticker.BIG_LIKE, sticker.BIG_LIKE),
                          ])
@pytest.mark.asyncio
async def test_match_different_size_of_fblike(match_size, message_size):
    trigger_any = utils.SimpleTrigger()
    trigger_likes = {size: utils.SimpleTrigger() for size in list(sticker.LIKE_STICKERS)}

    with answer.Talk() as talk:
        story = talk.story

        @story.on(sticker.Like(sticker.SMALL_LIKE))
        def small_like_sticker_story():
            @story.part()
            def then(ctx):
                trigger_likes[sticker.SMALL_LIKE].passed()

        @story.on(sticker.Like(sticker.MEDIUM_LIKE))
        def medium_like_sticker_story():
            @story.part()
            def then(ctx):
                trigger_likes[sticker.MEDIUM_LIKE].passed()

        @story.on(sticker.Like(sticker.BIG_LIKE))
        def big_like_sticker_story():
            @story.part()
            def then(ctx):
                trigger_likes[sticker.BIG_LIKE].passed()

        @story.on(sticker.Any())
        def any_like_sticker_story():
            @story.part()
            def then(ctx):
                trigger_any.passed()

        await talk.ask({
            'sticker_id': message_size,
        })

        assert trigger_likes[match_size].is_triggered



matchers = {}


def build_serializer():
    def default_serialize(_):
        pass

    return default_serialize


def build_deserialize(cls):
    def default_deserialize(data):
        try:
            return cls(data)
        except TypeError:
            try:
                return cls(**data)
            except TypeError:
                return cls()

    return default_deserialize


def get_validator(filter_data):
    """
    ask every matcher whether it can serve such filter data

    :param filter_data:
    :return:
    """
    for matcher_type, m in matchers.items():
        if hasattr(m, 'can_handle') and m.can_handle(filter_data):
            filter_data = m.handle(filter_data)

    return filter_data


def matcher():
    def register(m):
        m.type = getattr(m, 'type', m.__name__)
        if not getattr(m, 'serialize', False):
            m.serialize = build_serializer()
        if not getattr(m, 'deserialize', False):
            m.deserialize = build_deserialize(m)
        matchers[m.type] = m
        return m

    return register


def serialize(m):
    return {
        'type': m.type,
        'data': m.serialize(),
    }


def deserialize(data):
    matcher_type = matchers[data['type']]
    return matcher_type.deserialize(data['data'])



import logging
import pytest
import re
from . import text
from ... import matchers
from ...utils import answer, SimpleTrigger

logger = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_should_run_story_on_equal_message():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('hi there!')
        def one_story():
            @story.part()
            def then(ctx):
                trigger.passed()

        await talk.pure_text('hi there!')

        assert trigger.is_triggered


@pytest.mark.asyncio
async def test_should_not_run_story_on_non_equal_message():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('hi there!')
        def one_story():
            @story.part()
            def then(ctx):
                trigger.passed()

        await talk.pure_text('buy!')

        assert not trigger.is_triggered


@pytest.mark.asyncio
async def test_should_catch_any_text_message():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(text.Any())
        def one_story():
            @story.part()
            def then(ctx):
                trigger.passed()

        await talk.pure_text('hi there!')

        assert trigger.is_triggered


@pytest.mark.asyncio
async def test_should_ignore_any_non_text_message():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(text.Any())
        def one_story():
            @story.part()
            def then(ctx):
                trigger.passed()

        await talk.location('some where')

        assert not trigger.is_triggered


def test_serialize_text_any():
    m_old = text.Any()
    m_new = matchers.deserialize(matchers.serialize(m_old))
    assert isinstance(m_new, text.Any)


@pytest.mark.asyncio
async def test_should_catch_equal_text_message():
    trigger_hi_there = SimpleTrigger()
    trigger_see_you = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(text.Equal('hi there!'))
        def first_story():
            @story.part()
            def then(ctx):
                trigger_hi_there.passed()

        @story.on(text.Equal('see you!'))
        def second_story():
            @story.part()
            def then(ctx):
                trigger_see_you.passed()

        await talk.pure_text('see you!')

        assert not trigger_hi_there.is_triggered
        assert trigger_see_you.is_triggered


def test_equal_handle_should_create_right_type():
    assert isinstance(text.Equal.handle(''), text.Equal)


def test_serialize_text_equal():
    m_old = text.Equal('hats off')
    m_new = matchers.deserialize(matchers.serialize(m_old))
    assert isinstance(m_new, text.Equal)
    assert m_new.test_string == 'hats off'


@pytest.mark.asyncio
async def test_should_catch_equal_text_message_case_in_sensitive():
    trigger_hi_there = SimpleTrigger()
    trigger_see_you = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(text.EqualCaseIgnore('hi there!'))
        def first_story():
            @story.part()
            def then(ctx):
                trigger_hi_there.passed()

        @story.on(text.EqualCaseIgnore('see you!'))
        def second_story():
            @story.part()
            def then(ctx):
                trigger_see_you.passed()

        await talk.pure_text('See You!')

        assert not trigger_hi_there.is_triggered
        assert trigger_see_you.is_triggered


def test_serialize_text_equal_case_ignore():
    m_old = text.EqualCaseIgnore('hats off')
    m_new = matchers.deserialize(matchers.serialize(m_old))
    assert isinstance(m_new, text.EqualCaseIgnore)
    assert m_new.test_string == 'hats off'


@pytest.mark.asyncio
async def test_should_catch_text_message_that_match_regex():
    trigger_buy = SimpleTrigger()
    trigger_sell = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(text.Match('buy (.*)btc'))
        def one_story():
            @story.part()
            def then(ctx):
                trigger_buy.receive(text.get_text(ctx)['matches'][0])

        @story.on(text.Match('sell (.*)btc'))
        def another_story():
            @story.part()
            def then(ctx):
                trigger_sell.receive(text.get_text(ctx)['matches'][0])

        await talk.pure_text('buy 700btc')
        await talk.pure_text('sell 600btc')

        assert trigger_buy.result() == '700'
        assert trigger_sell.result() == '600'


@pytest.mark.asyncio
async def test_should_catch_text_message_that_match_regex_with_flags():
    trigger_destination = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(text.Match('going to (.*)', re.IGNORECASE))
        def one_story():
            @story.part()
            def then(ctx):
                logger.debug('ctx')
                logger.debug(ctx)
                trigger_destination.receive(text.get_text(ctx)['matches'][0])

        await talk.pure_text('Going to Pripyat')

        assert trigger_destination.result() == 'Pripyat'


@pytest.mark.asyncio
async def test_should_not_fail_on_empty_message():
    with answer.Talk() as talk:
        story = talk.story

        @story.on(text.Match('going to (.*)', re.IGNORECASE))
        def one_story():
            @story.part()
            def then(ctx):
                pass

        await talk.ask(None)


def test_serialize_text_match():
    m_old = text.Match('hello (.*)', re.IGNORECASE)
    m_new = matchers.deserialize(matchers.serialize(m_old))
    assert isinstance(m_new, text.Match)
    assert m_new.matcher.match('Hello Piter!')


def test_text_qual_should_handle_text():
    assert isinstance(matchers.get_validator('just pure text'), text.Equal)



from botstory import ast, matchers
import logging
import json

logger = logging.getLogger(__name__)


class Undefined:
    """
    Because we can got ever None value we should have something
    that definitely noted that value wasn't defined
    """

    def __init__(self):
        pass


class StoryPartFork:
    def __init__(self):
        self.local_scope = ast.library.StoriesScope()

    @property
    def __name__(self):
        return 'StoryPartFork'

    # this one very similar to validation in story loop
    def by_topic(self, topic):
        stories = self.local_scope.by_topic(topic)
        return stories[0] if len(stories) > 0 else None

    @property
    def children(self):
        return self.local_scope.stories

    def children_matcher(self):
        return Switch(
            self.local_scope.all_filters()
        )

    def should_loop(self):
        return False

    def __call__(self, *args, **kwargs):
        return None

    def get_child_by_validation_result(self, validation_result):
        case_stories = self.local_scope.get_story_by(case_id=validation_result)
        if len(case_stories) == 0:
            case_stories = self.local_scope.get_story_by(case_equal=validation_result)
        if len(case_stories) == 0:
            case_stories = self.local_scope.get_story_by(default_case=True)

        if len(case_stories) == 0:
            logger.debug('#######################################')
            logger.debug('# get_child_by_validation_result      #')
            logger.debug('#######################################')
            logger.debug('# [!] do not have any fork here       #')
            logger.debug('# context = {}'.format(self))
            logger.debug('# validation_result = {}'.format(validation_result))
            logger.debug('#######################################')
            return None

        return case_stories[0]

    def to_json(self):
        return {
            'type': 'StoryPartFork',
            'children': self.local_scope.to_json(),
        }

    def __repr__(self):
        return json.dumps(self.to_json())


@matchers.matcher()
class Switch:
    def __init__(self, cases):
        self.cases = cases

    def validate(self, message):
        for case_id, validator in self.cases.items():
            if validator.validate(message):
                return case_id
        return None

    def serialize(self):
        return [{
                    'id': id,
                    'data': matchers.serialize(c),
                } for id, c in self.cases.items()]

    @staticmethod
    def deserialize(data):
        return Switch({
                          case['id']: matchers.deserialize(case['data'])
                          for case in data
                          })

    def to_json(self):
        return {
            'type': 'Switch',
            'cases': self.serialize(),
        }

    def __repr__(self):
        return json.dumps(self.to_json())


class SwitchOnValue:
    """
    don't need to wait result
    """

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return 'SwitchOnValue({})'.format(self.value)


class ForkingStoriesAPI:
    def __init__(self, parser_instance):
        self.parser_instance = parser_instance

    def case(self,
             value=Undefined,
             validator=Undefined,
             default=Undefined,
             equal_to=Undefined,
             match=Undefined,
             ):
        def decorate(story_part):
            fork_node = self.parser_instance.get_last_story_part()
            if not isinstance(fork_node, StoryPartFork):
                fork_node = StoryPartFork()
                self.parser_instance.add_to_current_node(fork_node)

            compiled_story = self.parser_instance.compile_fork(fork_node, story_part)

            if default is True:
                compiled_story.extensions['default_case'] = True
            if equal_to is not Undefined:
                compiled_story.extensions['case_equal'] = equal_to
            if match is not Undefined:
                compiled_story.extensions['case_id'] = match
            if validator is not Undefined:
                compiled_story.extensions['validator'] = matchers.get_validator(validator)
            if value is not Undefined:
                compiled_story.extensions['validator'] = matchers.get_validator(value)
                compiled_story.extensions['case_equal'] = value

            return story_part

        return decorate



from botstory import matchers
from botstory.ast import story_context


def get_location(ctx):
    return story_context.get_message_data(ctx, 'location')


@matchers.matcher()
class Any:
    type = 'Location.Any'

    def __init__(self):
        pass

    def validate(self, ctx):
        return get_location(ctx)



import pytest
from .. import di


def test_inject_decorator():
    with di.child_scope():
        @di.inject()
        class OneClass:
            def __init__(self):
                pass

        assert isinstance(di.injector.get('one_class'), OneClass)


def test_bind_singleton_instance_by_default():
    with di.child_scope():
        @di.inject()
        class OneClass:
            def __init__(self):
                pass

        assert di.injector.get('one_class') == di.injector.get('one_class')


def test_inject_into_method_of_class():
    with di.child_scope():
        @di.inject()
        class OuterClass:
            @di.inject()
            def inner(self, inner_class):
                self.inner_class = inner_class

        @di.inject()
        class InnerClass:
            pass

        outer = di.injector.get('outer_class')
        assert isinstance(outer, OuterClass)
        assert isinstance(outer.inner_class, InnerClass)


def test_bind_should_inject_deps_in_decorated_methods_():
    with di.child_scope():
        @di.inject()
        class OuterClass:
            @di.inject()
            def inner(self, inner_class):
                self.inner_class = inner_class

        @di.inject()
        class InnerClass:
            pass

        outer = di.bind(OuterClass())
        assert isinstance(outer, OuterClass)
        assert isinstance(outer.inner_class, InnerClass)


def test_inject_default_value_if_we_dont_have_dep():
    with di.child_scope():
        @di.inject()
        class OuterClass:
            @di.inject()
            def inner(self, inner_class='Hello World!'):
                self.inner_class = inner_class

        outer = di.bind(OuterClass())
        assert isinstance(outer, OuterClass)
        assert outer.inner_class == 'Hello World!'


def test_no_autoupdate_deps_on_new_instance_comes():
    with di.child_scope():
        @di.inject()
        class OuterClass:
            @di.inject()
            def inner(self, inner_class=None):
                self.inner_class = inner_class

        outer = di.bind(OuterClass(), auto=False)

        @di.inject()
        class InnerClass:
            pass

        assert isinstance(outer, OuterClass)
        assert outer.inner_class is None


def test_autoupdate_deps_on_new_instance_comes():
    with di.child_scope():
        @di.inject()
        class OuterClass:
            @di.inject()
            def inner(self, inner_class=None):
                self.inner_class = inner_class

        outer = di.bind(OuterClass(), auto=True)

        @di.inject()
        class InnerClass:
            pass

        assert isinstance(outer, OuterClass)
        assert isinstance(outer.inner_class, InnerClass)


def test_fail_on_cyclic_deps():
    with di.child_scope():
        @di.inject()
        class FirstClass:
            @di.inject()
            def deps(self, second_class=None):
                self.second_class = second_class

        @di.inject()
        class SecondClass:
            @di.inject()
            def deps(self, first_class=None):
                self.first_class = first_class

        first_class = di.injector.get('first_class')
        assert isinstance(first_class.second_class, SecondClass)
        assert isinstance(first_class.second_class.first_class, FirstClass)


def test_custom_type():
    with di.child_scope():
        @di.inject('qwerty')
        class OneClass:
            pass

        assert isinstance(di.injector.get('qwerty'), OneClass)


def test_fail_on_incorrect_using():
    with pytest.raises(NotImplementedError):
        di.inject()('qwerty')



import aiohttp
import asyncio
from botstory.ast import story_context
import json
import pytest
from unittest import mock

from . import GAStatistics, tracker
from .. import fb, ga, mockdb, mockhttp
from ... import di, Story, utils

story = None


def teardown_function(function):
    story and story.clear()


@pytest.fixture
def tracker_mock(mocker):
    tr = mock.Mock()
    tr.send = aiohttp.test_utils.make_mocked_coro()
    mocker.patch.object(tracker, 'Tracker', return_value=tr)
    return tr


@pytest.mark.asyncio
async def test_should_put_in_queue_story_tracker(mocker, tracker_mock):
    user = utils.build_fake_user()
    ga = GAStatistics(tracking_id='UA-XXXXX-Y')

    class FakePart:
        @property
        def __name__(self):
            return 'one part'

    ctx = story_context.StoryContext(None, None)

    with mock.patch.object(ctx,
                           'user',
                           return_value=user):
        with mock.patch.object(ctx,
                               'stack',
                               return_value=[{'topic': 'one story'}]):
            with mock.patch.object(ctx,
                                   'get_current_story_part',
                                   return_value=FakePart()):
                ga.story(ctx)

    await asyncio.sleep(0.1)

    tracker_mock.send.assert_called_once_with(
        'pageview',
        'one story/one part',
    )


@pytest.mark.asyncio
async def test_should_put_in_queue_new_message_tracker(tracker_mock):
    user = utils.build_fake_user()
    ga = GAStatistics(tracking_id='UA-XXXXX-Y')

    ga.new_message(
        story_context.StoryContext(
            {
                'user': user,
                'session': {'data': {'text': {'raw': 'hi!'}}},
            }, None
        )
    )

    await asyncio.sleep(0.1)

    tracker_mock.send.assert_called_once_with(
        'pageview',
        'receive: {}'.format(json.dumps({'text': {'raw': 'hi!'}})),
    )


@pytest.mark.asyncio
async def test_should_put_in_queue_new_user_tracker(tracker_mock):
    user = utils.build_fake_user()
    ga = GAStatistics(tracking_id='UA-XXXXX-Y')

    ga.new_user(user)

    await asyncio.sleep(0.1)

    tracker_mock.send.assert_called_once_with(
        'event',
        'new_user', 'start', 'new user starts chat'
    )


@pytest.mark.asyncio
async def test_should_put_in_queue_event_tracker(tracker_mock):
    user = utils.build_fake_user()
    ga = GAStatistics(tracking_id='UA-XXXXX-Y')

    ga.event(user, 'category', 'action', 'label', 10)

    await asyncio.sleep(0.1)

    tracker_mock.send.assert_called_once_with(
        'event',
        'category', 'action', 'label', 10,
    )


# integration testing

@pytest.mark.asyncio
async def test_should_track_story(tracker_mock):
    global story
    story = Story()

    @story.on('hi!')
    def one_story():
        @story.part()
        def greeting(message):
            pass

    story.use(mockdb.MockDB())
    facebook = story.use(fb.FBInterface())
    story.use(mockhttp.MockHttpInterface())
    story.use(ga.GAStatistics(tracking_id='UA-XXXXX-Y'))
    await story.start()

    await facebook.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 123456789,
            'messaging': [{
                'sender': {
                    'id': 'UNKNOWN_ID',
                },
                'recipient': {
                    'id': 'PAGE_ID'
                },
                'timestamp': 123456789,
                'message': {
                    'text': 'hi!'
                }
            }]
        }]
    })

    await asyncio.sleep(0.1)
    tracker_mock.send.assert_has_calls([
        # mock.call('event',
        #           'new_user', 'start', 'new user starts chat'
        #           ),
        # mock.call('pageview',
        #           'receive: {}'.format(json.dumps({'text': {'raw': 'hi!'}})),
        #           ),
        mock.call('pageview',
                  'one_story/greeting',
                  ),
    ])


def test_get_as_deps():
    global story
    story = Story()

    story.use(ga.GAStatistics())

    with di.child_scope():
        @di.desc()
        class OneClass:
            @di.inject()
            def deps(self, tracker):
                self.tracker = tracker

        assert isinstance(di.injector.get('one_class').tracker, ga.GAStatistics)



from botstory import utils


def test_safe_get_existing_property():
    assert utils.safe_get({
        'a': {
            'b': {
                'c': 'Hello World'
            }
        }
    }, 'a', 'b', 'c') == 'Hello World'


def test_safe_get_nonexistent_property():
    assert utils.safe_get({
        'a': {
            'b': {
                'c': 'Hello World'
            }
        }
    }, 'c', 'b', 'a') is None


def test_safe_get_fail_to_default():
    assert utils.safe_get({
        'a': {
            'b': {
                'c': 'Hello World'
            }
        }
    }, 'c', 'b', 'a', default='default-value') == 'default-value'


def test_safe_get_of_none_is_none():
    assert utils.safe_get(None, 'c', 'b', 'a') is None


def test_safe_get_of_dict_with_none():
    assert utils.safe_get({
        'c': None
    }, 'c', 'b', 'a') is None


def test_safe_set_existing_property():
    assert utils.safe_set({
        'a': {
            'b': {
                'c': 'Mercury',
            }
        }
    }, 'a', 'b', 'c', 'Venus') == {
               'a': {
                   'b': {
                       'c': 'Venus',
                   }
               }
           }


def test_safe_set_nonexistent_property():
    assert utils.safe_set({
        'a': {
            'b': {
                'c': 'Mercury',
            }
        }
    }, 'a', 'b2', 'c2', 'Venus') == {
               'a': {
                   'b': {
                       'c': 'Mercury',
                   },
                   'b2': {
                       'c2': 'Venus',
                   }
               }
           }


def test_safe_set_updates_by_dict():
    assert utils.safe_set({
        'planets': {
            'Mercury': {
                'gravity': 3.7,
                'mass': 0.055,
                'radius': 2440,
            },
            'Venus': {
                'gravity': 8.87,
                'mass': 0.815,
                'radius': 6052,
            },
            'Earth': {
                'error': 'missed',
            }
        }
    }, 'planets', 'Earth', {
        'gravity': 9.8,
        'radius': 6371,
        'mass': 1,
    }) == {
           'planets': {
               'Mercury': {
                   'gravity': 3.7,
                   'mass': 0.055,
                   'radius': 2440,
               },
               'Venus': {
                   'gravity': 8.87,
                   'mass': 0.815,
                   'radius': 6052,
               },
               'Earth': {
                   'gravity': 9.8,
                   'radius': 6371,
                   'mass': 1,
               }
           }
       }



import asyncio
import logging

from . import chat, di
from .ast import callable as callable_module, common, \
    forking, library, loop, parser, processor, users

logger = logging.getLogger(__name__)


def check_spec(spec, obj):
    for method in [getattr(obj, s, None) for s in spec]:
        if not method:
            return False
    return True


class Story:
    def __init__(self):
        self.stories_library = library.StoriesLibrary()

        self.parser_instance = parser.Parser(self.stories_library)

        self.story_processor_instance = processor.StoryProcessor(
            self.parser_instance,
            self.stories_library,
        )

        self.match_message = self.story_processor_instance.match_message

        self.callable_stories_instance = callable_module.CallableStoriesAPI(
            library=self.stories_library,
            parser_instance=self.parser_instance,
            processor_instance=self.story_processor_instance,
        )
        self.common_stories_instance = common.CommonStoriesAPI(
            self.parser_instance,
            self.stories_library)
        self.forking_api = forking.ForkingStoriesAPI(
            parser_instance=self.parser_instance,
        )
        self.story_loop = loop.StoryLoopAPI(
            library=self.stories_library,
            parser_instance=self.parser_instance,
        )
        self.middlewares = []
        self.chat = chat.Chat()
        self.users = users.Users()

    # Facade
    def loop(self):
        """
        close loop/scope of sub-stories
        :return:
        """
        return self.story_loop.loop()

    def on(self, receive):
        return self.common_stories_instance.on(receive)

    def on_start(self):
        return self.common_stories_instance.on_start()

    def part(self):
        return self.common_stories_instance.part()

    def callable(self):
        return self.callable_stories_instance.callable()

    def case(self,
             value=forking.Undefined,
             *,
             validator=forking.Undefined,
             default=forking.Undefined,
             equal_to=forking.Undefined,
             match=forking.Undefined
             ):
        return self.forking_api.case(
            value,
            validator,
            default,
            equal_to,
            match,
        )

    async def ask(self, body, quick_replies=None, options=None, user=None):
        return await self.chat.ask(body, quick_replies, options, user)

    async def say(self, body, user, options=None):
        return await self.chat.say(body, user, options)

    def use(self, middleware):
        """
        attache middleware

        :param middleware:
        :return:
        """

        logger.debug('use')
        logger.debug(middleware)

        self.middlewares.append(middleware)

        di.injector.register(instance=middleware)
        di.bind(middleware, auto=True)

        # TODO: should use DI somehow
        if check_spec(['send_text_message'], middleware):
            self.chat.add_interface(middleware)

        return middleware

    async def setup(self, event_loop=None):
        self.register()
        return await self._do_for_each_extension('setup', event_loop)

    async def start(self, event_loop=None):
        self.register()
        await self._do_for_each_extension('before_start', event_loop)
        await self._do_for_each_extension('start', event_loop)
        await self._do_for_each_extension('after_start', event_loop)

    async def stop(self, event_loop=None):
        return await self._do_for_each_extension('stop', event_loop)

    def forever(self, loop):
        try:
            loop.run_forever()
        except KeyboardInterrupt:  # pragma: no cover
            pass
        finally:
            loop.run_until_complete(self.stop())

    def register(self):
        di.injector.register(instance=self.parser_instance)
        di.injector.register(instance=self.story_processor_instance)
        di.injector.register(instance=self.stories_library)
        di.injector.register(instance=self.users)
        di.injector.bind(self.parser_instance, auto=True)
        di.injector.bind(self.story_processor_instance, auto=True)
        di.injector.bind(self.stories_library, auto=True)
        di.injector.bind(self.users, auto=True)

    async def _do_for_each_extension(self, command, even_loop):
        await asyncio.gather(
            *[getattr(m, command)() for m in self.middlewares if hasattr(m, command)],
            loop=even_loop)

    def clear(self):
        """
        Clear all deps
        TODO: replace with DI

        :return:
        """

        di.clear_instances()



from botstory.utils import advanced_json_encoder
from bson import ObjectId
import json


def test_encode_objectid():
    assert advanced_json_encoder.AdvancedJSONEncoder().encode({
        'id': ObjectId('51948e86c25f4b1d1c0d303c'),
        'message': 'hello world!',
    }) == json.dumps({
        'id': '51948e86c25f4b1d1c0d303c',
        'message': 'hello world!',
    })



import logging
from ... import di

logger = logging.getLogger(__name__)


@di.desc('tracker', reg=False)
class MockTracker:
    def event(self, *args, **kwargs):
        logging.debug('event')
        logging.debug(kwargs)
        logging.debug(args)

    def new_message(self, *args, **kwargs):
        logging.debug('new_message')
        logging.debug(kwargs)
        logging.debug(args)

    def new_user(self, *args, **kwargs):
        logging.debug('new_user')
        logging.debug(kwargs)
        logging.debug(args)

    def story(self, *args, **kwargs):
        logging.debug('story')
        logging.debug(kwargs)
        logging.debug(args)



from botstory import di
from botstory.ast import story_context
from botstory.integrations import mocktracker

import logging

logger = logging.getLogger(__name__)


@di.desc(reg=False)
class StoryProcessor:
    def __init__(self, parser_instance, library):
        self.library = library
        self.parser_instance = parser_instance
        self.tracker = mocktracker.MockTracker()

    @di.inject()
    def add_tracker(self, tracker):
        logger.debug('add_tracker')
        logger.debug(tracker)
        if not tracker:
            return
        self.tracker = tracker

    async def match_message(self, message_ctx):
        """

        match bot message to existing stories
        and take into account context of current user

        :param message_ctx:
        :return: mutated message_ctx
        """
        logger.debug('')
        logger.debug('# match_message')
        logger.debug('')
        logger.debug(message_ctx)

        ctx = story_context.StoryContext(library=self.library,
                                         matched=False,
                                         message=message_ctx,
                                         )

        self.tracker.new_message(ctx)

        if ctx.is_empty_stack():
            if not ctx.does_it_match_any_story():
                # there is no stories for such message
                return ctx.message

            ctx = story_context.reducers.scope_in(ctx)
            ctx = await self.process_story(ctx)
            ctx = story_context.reducers.scope_out(ctx)

        while ctx.could_scope_out() and not ctx.is_empty_stack():
            logger.debug('# in a loop')
            logger.debug(ctx)

            # looking for first valid matcher
            while True:
                if ctx.is_empty_stack():
                    # we have reach the bottom of stack
                    if ctx.does_it_match_any_story():
                        # but sometimes we could jump on other global matcher
                        ctx = story_context.reducers.scope_in(ctx)
                        ctx = await self.process_story(ctx)
                        ctx = story_context.reducers.scope_out(ctx)
                        if ctx.is_empty_stack():
                            return ctx.message
                    else:
                        logger.debug('  we have reach the bottom of stack '
                                     'so no once has receive this message')
                        return ctx.message

                if ctx.is_scope_level() and \
                        (ctx.has_child_story() or ctx.matched) and \
                        not ctx.is_breaking_a_loop() or \
                                not ctx.is_scope_level() and \
                                not ctx.is_end_of_story():
                    # if we
                    # - haven't reach last step in list of story
                    # - isn't breaking a loop
                    # - inside of child scope and have matching sub-stories
                    # so we can parse result
                    break

                logger.debug('# tuning scope_out')
                ctx = story_context.reducers.scope_out(ctx)

            if ctx.has_child_story():
                ctx = story_context.reducers.scope_in(ctx)
            ctx = await self.process_story(ctx)
            logger.debug('# match_message scope_out')
            ctx = story_context.reducers.scope_out(ctx)

            if ctx.is_scope_level() and not ctx.is_breaking_a_loop():
                logger.debug('# the end of one story scope')
                break

        return ctx.message

    async def process_story(self, ctx):
        logger.debug('')
        logger.debug('# process_story')
        logger.debug('')
        logger.debug(ctx)

        if ctx.is_scope_level():
            return story_context.reducers.enter_new_scope(ctx)

        # integrate over parts of story
        last_story_part = None
        storyline = story_context.reducers.iterate_storyline(ctx)
        if not storyline:
            return ctx
        try:
            story_part_ctx = next(storyline)

            # loop thought storyline
            # we can't use for here because we should send update context
            # back to iterator
            while True:
                self.tracker.story(story_part_ctx)

                if story_part_ctx.has_child_story() or story_part_ctx.is_scope_level_part():
                    story_part_ctx = story_context.reducers.scope_in(story_part_ctx)
                    story_part_ctx = await self.process_story(story_part_ctx)
                    logger.debug('# process_story scope_out')
                    story_part_ctx = story_context.reducers.scope_out(story_part_ctx)
                else:
                    story_part_ctx = await story_context.reducers.execute(story_part_ctx)

                if story_part_ctx.is_waiting_for_input():
                    return story_part_ctx
                last_story_part = story_part_ctx
                story_part_ctx = storyline.send(story_part_ctx)
        except StopIteration:
            pass

        if last_story_part is None:
            logger.warn('# story line is empty or it is differ from what we have in stack')
            return ctx
        return last_story_part



from botstory.ast import story_context


def test_clean_message_data():
    ctx = {
        'session': {
            'data': {
                'message': {
                    'text': {
                        'raw': 'I\'m going to Mars',
                        'something': 'else',
                    },
                    'location': {
                        'lat': 12,
                        'lng': 21,
                    }
                }
            }
        }
    }
    ctx = story_context.clean_message_data(ctx)
    assert ctx == {
        'session': {
            'data': {
                'message': {
                }
            }
        }
    }


def test_get_message_data():
    assert story_context.get_message_data({
        'session': {
            'data': {
                'message': {
                    'text': {
                        'raw': 'I\'m going to Mars',
                    },
                    'location': {
                        'lat': 12,
                        'lng': 21,
                    }
                }
            }
        }
    }, 'text') == {
               'raw': 'I\'m going to Mars',
           }


def test_set_message_data():
    ctx = {
        'session': {
            'data': {
                'message': {
                    'text': {
                        'raw': 'I\'m going to Mars',
                        'something': 'else',
                    },
                    'location': {
                        'lat': 12,
                        'lng': 21,
                    }
                }
            }
        }
    }
    ctx = story_context.set_message_data(ctx, 'text', {
        'raw': 'I\'m going to Venus'
    })
    assert ctx == {
        'session': {
            'data': {
                'message': {
                    'text': {
                        'raw': 'I\'m going to Venus',
                    },
                    'location': {
                        'lat': 12,
                        'lng': 21,
                    }
                }
            }
        }
    }



from aiohttp import test_utils
import json
import pytest
from . import AioHttpInterface
from .. import aiohttp
from ..commonhttp import errors
from ..tests import fake_server
from ... import di, Story

story = None


def teardown_function(function):
    story and story.clear()


@pytest.fixture
def webhook_handler():
    return test_utils.make_mocked_coro(return_value={
        'status': 200,
        'content_type': 'application/json',
        'text': json.dumps({'message': 'Ok!'}),
    })


@pytest.mark.asyncio
async def test_post(event_loop):
    async with fake_server.FakeFacebook(event_loop) as server:
        async with server.session() as session:
            http = AioHttpInterface()
            http.session = session

            assert await http.post(fake_server.URI.format('/v2.6/me/messages/'), json={'message': 'hello world!'})
            assert len(server.history) == 1
            req = server.history[-1]['request']
            assert req.content_type == 'application/json'
            assert await req.json() == {'message': 'hello world!'}


@pytest.mark.asyncio
async def test_listen_webhook(webhook_handler):
    http = AioHttpInterface(port=9876)
    http.webhook(uri='/webhook', handler=webhook_handler, token='qwerty')
    try:
        await http.start()
        res = await http.post_raw('http://localhost:9876/webhook', json={'message': 'Is there anybody in there?'})
        webhook_handler.assert_called_once_with({'message': 'Is there anybody in there?'})
        assert res['status'] == 200
        assert res['text'] == json.dumps({'message': 'Ok!'})
    finally:
        await http.stop()


@pytest.mark.asyncio
async def test_pass_validation_for_correct_request():
    http = AioHttpInterface(port=9876)
    http.webhook(uri='/webhook', handler=None, token='token-value')
    try:
        await http.start()
        res = await http.get_raw('http://localhost:9876/webhook', params={
            'hub.challenge': 'some-challenge',
            'hub.mode': 'subscribe',
            'hub.verify_token': 'token-value',
        })
        assert res['text'] == 'some-challenge'
    finally:
        await http.stop()


@pytest.mark.asyncio
async def test_reject_validation_for_incorrect_request():
    http = AioHttpInterface(port=9876)
    http.webhook(uri='/webhook', handler=None, token='token-value')
    try:
        await http.start()
        await http.get('http://localhost:9876/webhook', params={
            'something': 'incorrect',
        })
    except errors.HttpRequestError as err:
        assert err.code == 422
        assert err.message == 'Error, wrong validation token'
    finally:
        await http.stop()


@pytest.mark.asyncio
async def test_should_not_create_server_if_there_wasnt_any_webhooks():
    http = AioHttpInterface(port=9876)
    try:
        await http.start()
        assert not http.server
    finally:
        await http.stop()


@pytest.mark.skip(reason='too long')
@pytest.mark.asynciod
async def test_get_400_from_wrong_domain():
    http = AioHttpInterface()
    try:
        await http.get('http://wrong-url')
    except errors.HttpRequestError as err:
        assert err.code == 400


@pytest.mark.asyncio
async def test_get_400_from_wrong_path():
    http = AioHttpInterface()
    try:
        await http.get('http://localhost:9876/webhook')
    except errors.HttpRequestError as err:
        assert err.code == 400


@pytest.mark.asyncio
async def test_post_to_wrong_path_get_400(event_loop):
    async with fake_server.FakeFacebook(event_loop) as server:
        async with server.session() as session:
            http = AioHttpInterface()
            http.session = session

            try:
                await http.post(fake_server.URI.format('/v2.6/me/messages/mistake'), json={'message': 'hello world!'})
            except errors.HttpRequestError as err:
                assert err.code == 404


@pytest.mark.skip(reason='too long')
@pytest.mark.asyncio
async def test_post_400_from_wrong_domain():
    http = AioHttpInterface()
    try:
        await http.post('http://wrong-url')
    except errors.HttpRequestError as err:
        assert err.code == 400


@pytest.mark.asyncio
async def test_post_400_from_wrong_path():
    http = AioHttpInterface()
    try:
        await http.post('http://localhost:9876/webhook')
    except errors.HttpRequestError as err:
        assert err.code == 400


@pytest.mark.asyncio
async def test_delete_200(event_loop):
    async with fake_server.FakeFacebook(event_loop) as server:
        async with server.session() as session:
            http = AioHttpInterface()
            http.session = session

            await http.delete(fake_server.URI.format('/v2.6/me/thread_settings'))


@pytest.mark.asyncio
async def test_delete_400(event_loop):
    async with fake_server.FakeFacebook(event_loop) as server:
        async with server.session() as session:
            http = AioHttpInterface()
            http.session = session

            try:
                await http.delete(fake_server.URI.format('/v2.6/me/wrong'))
            except errors.HttpRequestError as err:
                assert err.code == 404


@pytest.mark.asyncio
async def test_pass_middleware(mocker, webhook_handler):
    handler_stub = mocker.stub()

    def MockMiddleware():
        async def middleware_factory(app, handler):
            async def mock_middleware_handler(request):
                handler_stub(request)
                return await handler(request)

            return mock_middleware_handler

        return middleware_factory

    try:
        http = AioHttpInterface(middlewares=[MockMiddleware()], port=8080)
        http.webhook(uri='/webhook', handler=webhook_handler, token='qwerty')
        await http.start()
        await http.post('http://localhost:8080/webhook', json={})
        assert handler_stub.called
    finally:
        await http.stop()


def test_get_as_deps():
    global story
    story = Story()
    story.use(aiohttp.AioHttpInterface())

    with di.child_scope('http'):
        @di.desc()
        class OneClass:
            @di.inject()
            def deps(self, http):
                self.http = http

        assert isinstance(di.injector.get('one_class').http, aiohttp.AioHttpInterface)


@pytest.mark.asyncio
async def test_shop_should_remove_app():
    http = aiohttp.AioHttpInterface()
    assert not http.has_app()
    http.get_app()
    assert http.has_app()
    await http.stop()
    assert not http.has_app()


@pytest.mark.asyncio
async def test_should_raise_exception_on_webhook_if_aiohttp_already_is_started(webhook_handler):
    http = aiohttp.AioHttpInterface()
    http.get_app()
    await http.start()
    with pytest.raises(aiohttp.WebhookException):
        http.webhook(uri='/webhook', handler=webhook_handler, token='qwerty')



import aiohttp
from aiohttp import errors, web
import asyncio
import logging
import json as _json
import urllib
from yarl import URL

from ..commonhttp import errors as common_errors, statuses
from ... import di

logger = logging.getLogger(__name__)


class WebhookHandler:
    def __init__(self, handler):
        self.handler = handler

    async def handle(self, request):
        res = await self.handler(await request.json())
        return web.Response(**res)


def is_ok(status):
    return 200 <= status < 400


class WebhookException(Exception):
    pass


@di.desc('http', reg=False)
class AioHttpInterface:
    def __init__(self, host='0.0.0.0', port=None,
                 shutdown_timeout=60.0, ssl_context=None,
                 backlog=128, auto_start=True,
                 middlewares=[],
                 ):
        if port is None:
            if not ssl_context:
                port = 8080
            else:
                port = 8443

        self.backlog = backlog
        self.host = host
        self.middlewares = middlewares
        self.port = port
        self.shutdown_timeout = shutdown_timeout

        self.ssl_context = ssl_context
        self.auto_start = auto_start

        self.app = None
        self.session = None
        self.server = None
        self.handler = None
        self.webhook_token = None

    async def get(self, url, params=None, headers=None):
        logger.debug('get url={}'.format(url))
        loop = asyncio.get_event_loop()
        with aiohttp.ClientSession(loop=loop) as session:
            return await(await self.method(
                method_type='get',
                session=session,
                url=url,
                params=params,
                headers=headers,
            )).json()

    async def get_raw(self, url, params=None, headers=None):
        logger.debug('get url={}'.format(url))
        loop = asyncio.get_event_loop()
        with aiohttp.ClientSession(loop=loop) as session:
            res = await self.method(
                method_type='get',
                session=session,
                url=url,
                params=params,
                headers=headers,
            )
            return {
                'status': res.status,
                'headers': res.headers,
                'text': await res.text(),
            }

    async def post(self, url, params=None, headers=None, json=None):
        logger.debug('post url={}'.format(url))
        headers = headers or {}
        headers['Content-Type'] = headers.get('Content-Type', 'application/json')
        loop = asyncio.get_event_loop()
        with aiohttp.ClientSession(loop=loop) as session:
            return await(await self.method(
                method_type='post',
                session=session,
                url=url,
                params=params,
                headers=headers,
                data=_json.dumps(json),
            )).json()

    async def post_raw(self, url, params=None, headers=None, json=None):
        logger.debug('post url={}'.format(url))
        headers = headers or {}
        headers['Content-Type'] = headers.get('Content-Type', 'application/json')
        loop = asyncio.get_event_loop()
        with aiohttp.ClientSession(loop=loop) as session:
            res = await self.method(
                method_type='post',
                session=session,
                url=url,
                params=params,
                headers=headers,
                data=_json.dumps(json),
            )
            return {
                'status': res.status,
                'headers': res.headers,
                'text': await res.text(),
            }

    async def delete(self, url, params=None, headers=None, json=None):
        logger.debug('delete url={}'.format(url))
        headers = headers or {}
        headers['Content-Type'] = headers.get('Content-Type', 'application/json')
        loop = asyncio.get_event_loop()
        with aiohttp.ClientSession(loop=loop) as session:
            return await(await self.method(
                method_type='delete',
                session=session,
                url=url,
                params=params,
                headers=headers,
                data=_json.dumps(json),
            )).json()

    async def method(self, method_type, session, url, **kwargs):
        # be able to mock session from outside
        session = self.session or session
        try:
            try:
                method_name = getattr(session, method_type)
                resp = await method_name(
                    url,
                    **kwargs,
                )
            except errors.ClientOSError as err:
                raise common_errors.HttpRequestError(
                    code=400,
                    message='{} {}'.format(err.errno, err.strerror),
                )
            if not is_ok(resp.status):
                raise common_errors.HttpRequestError(
                    code=resp.status,
                    headers=resp.headers,
                    message=await resp.text(),
                )
        except Exception as err:
            logger.warn(
                'Exception: status: {status}, message: {message}, type: {type}, method: {method}, url: {url}, {kwargs}'
                    .format(status=getattr(err, 'code', None),
                            message=getattr(err, 'message', None),
                            type=type(err),
                            method=method_name,
                            url=url,
                            kwargs=kwargs,
                            )
            )
            raise err
        return resp

    def get_app(self):
        if not self.has_app():
            loop = asyncio.get_event_loop()
            logger.debug('create web app')
            self.app = web.Application(
                loop=loop,
                middlewares=self.middlewares,
            )
        return self.app

    def has_app(self):
        return self.app is not None

    def webhook(self, uri, handler, token):
        logger.debug('register webhook {}'.format(uri))
        self.webhook_token = token
        if self.get_app().frozen:
            raise WebhookException('Aiohttp extension is already started. '
                                   'We should change webhook before aiohttp is started.')
        self.get_app().router.add_get(uri, self.handle_webhook_validation)
        self.get_app().router.add_post(uri, WebhookHandler(handler).handle)

    def handle_webhook_validation(self, request):
        params = {name: value[0] for name, value in urllib.parse.parse_qs(request.query_string).items()}
        logger.debug('try to validate webhook with {}'.format(params))
        if params.get('hub.verify_token', None) == self.webhook_token and \
                        params.get('hub.mode', None) == 'subscribe':
            return web.Response(text=params.get('hub.challenge'))
        else:
            return web.Response(text='Error, wrong validation token',
                                status=statuses.HTTP_422_UNPROCESSABLE_ENTITY)

    async def start(self):
        logger.debug('start')
        if not self.has_app():
            logger.debug('does not have app')
            return
        if not self.auto_start:
            logger.debug('should not start automatically')
            return
        app = self.get_app()
        handler = app.make_handler()
        loop = asyncio.get_event_loop()
        server = loop.create_server(
            handler,
            self.host,
            self.port,
            ssl=self.ssl_context,
            backlog=self.backlog,
        )

        srv, startup_res = await asyncio.gather(
            server, app.startup(),
            loop=loop,
        )

        scheme = 'https' if self.ssl_context else 'http'
        url = URL('{}://localhost'.format(scheme))
        url = url.with_host(self.host).with_port(self.port)
        logger.debug('======== Running on {} ========\n'
                     '(Press CTRL+C to quit)'.format(url))

        self.server = srv
        self.handler = handler

    async def stop(self):
        logger.debug('stop')

        if not self.has_app():
            logger.debug('does not have app')
            return

        if self.server:
            self.server.close()
            await self.server.wait_closed()

        await self.app.shutdown()
        await self.app.cleanup()
        self.app = None

        if self.handler:
            await self.handler.finish_connections(self.shutdown_timeout)



import logging
import pytest
import random

from botstory.ast import callable, loop, story_context
from botstory.middlewares import text
from .. import EndOfStory, SwitchOnValue
from ..utils import answer, SimpleTrigger

logger = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_begin_of_callable_story():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def one_story():
            @story.part()
            def store_arguments(ctx):
                data = story_context.get_user_data(ctx)
                trigger.receive({
                    'value1': data['arg1'],
                    'value2': data['arg2'],
                })

        await one_story(arg1=1, arg2=2, session=talk.session, user=talk.user)

        assert trigger.result() == {
            'value1': 1,
            'value2': 2,
        }


@pytest.mark.asyncio
async def test_parts_of_callable_story():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def meet_ava_story():
            @story.part()
            async def ask_name(ctx):
                return await story.ask(
                    'My name is Ava. What is your name?',
                    user=ctx['user'],
                )

            @story.part()
            async def ask_age(ctx):
                trigger_1.passed()
                return await story.ask(
                    'Nice to see you {}. What do you do here?'.format(story_context.get_user_data(ctx)),
                    user=ctx['user'],
                )

            @story.part()
            async def store_arguments(ctx):
                age = int(text.get_raw_text(ctx))
                if age < 30:
                    res = 'You are so young! '
                else:
                    res = 'Hm. Too old to die young'

                await story.say(res, user=ctx['user'])
                trigger_2.passed()

        ctx = await meet_ava_story(session=talk.session, user=talk.user)
        talk.init_by_ctx(ctx)

        await talk.pure_text('Eugene')
        await talk.pure_text('13')

        assert trigger_1.is_triggered
        assert trigger_2.is_triggered


@pytest.mark.asyncio
async def test_call_story_from_common_story():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.callable()
        def common_greeting():
            @story.part()
            async def ask_name(ctx):
                return await story.ask(
                    'Hi {}. How are you?'.format(ctx['user']['name']),
                    user=ctx['user'],
                )

        @story.on('Hi!')
        def meet():
            @story.part()
            async def greeting(ctx):
                return await common_greeting(
                    user=ctx['user'],
                    session=ctx['session'],
                )

            @story.part()
            async def ask_location(ctx):
                return await story.ask(
                    'Which planet are we going to visit today?',
                    user=ctx['user'],
                )

            @story.part()
            def parse(ctx):
                trigger.receive(text.get_raw_text(ctx))

        await say_pure_text('Hi!')
        await say_pure_text('I\'m fine')
        await say_pure_text('Venus, as usual!')

        assert trigger.value == 'Venus, as usual!'


@pytest.mark.asyncio
async def test_parts_of_callable_story_can_be_sync():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def one_story():
            @story.part()
            def has(ctx):
                trigger_1.passed()

            @story.part()
            def so(ctx):
                trigger_2.passed()

        await one_story(session=talk.session, user=talk.user)

        assert trigger_1.is_triggered
        assert trigger_2.is_triggered


@pytest.mark.asyncio
async def test_call_story_from_another_callable():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def one_story():
            @story.part()
            def so_1(ctx):
                pass

            @story.part()
            async def so_2(ctx):
                await another_story(session=ctx['session'], user=ctx['user'])

            @story.part()
            def so_3(ctx):
                trigger_2.passed()

        @story.callable()
        def another_story():
            @story.part()
            def has(cxt):
                pass

            @story.part()
            def so(ctx):
                trigger_1.passed()
                return callable.EndOfStory()

        # push extra parameter with session
        # and it will propagate up to other story as well
        logger.debug('[!] before await one_story')
        await one_story(session=talk.session, user=talk.user)

        assert trigger_1.is_triggered
        assert trigger_2.is_triggered


@pytest.mark.asyncio
async def test_async_end_of_story_with_switch():
    sides = ['heads', 'tails']
    budget = SimpleTrigger(2)
    game_result = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def flip_a_coin():
            @story.part()
            async def choose_side(ctx):
                return await story.ask('Please choose a side', user=ctx['user'])

            @story.part()
            async def flip(ctx):
                user_side = text.get_raw_text(ctx)
                await story.say('Thanks!', user=ctx['user'])
                await story.say('And I am flipping a Coin', user=ctx['user'])
                coin_side = random.choice(sides)
                await story.say('and got {}'.format(coin_side), user=ctx['user'])
                if coin_side == user_side:
                    await story.say('My greetings! You guessed!', user=ctx['user'])
                    budget.receive(budget.value + 1)
                else:
                    await story.say('Sad but you loose!', user=ctx['user'])
                    budget.receive(budget.value - 1)
                return SwitchOnValue(budget.value)

            @story.case(equal_to=0)
            def loose():
                @story.part()
                def end_of_story(ctx):
                    return EndOfStory({
                        'game_result': 'loose'
                    })

            @story.case(equal_to=5)
            def win():
                @story.part()
                def end_of_story(ctx):
                    return EndOfStory({
                        'game_result': 'win'
                    })

            @story.part()
            def tail_recursion(ctx):
                # TODO: add test for recursion
                # return flip_a_coin(message['user'], session)
                return EndOfStory({
                    'game_result': 'in progress'
                })

        @story.on('enter to the saloon')
        def enter_to_the_saloon():
            @story.part()
            async def start_a_game(ctx):
                return await flip_a_coin(user=ctx['user'], session=ctx['session'])

            @story.part()
            def game_over(ctx):
                logger.debug('game_over')
                logger.debug(ctx)
                game_result.receive(story_context.get_user_data(ctx)['game_result'])

        await talk.pure_text('enter to the saloon')

        await talk.pure_text(random.choice(sides))

        assert game_result.result() in ['loose', 'win', 'in progress']


@pytest.mark.asyncio
@pytest.mark.skip('recursion')
async def test_async_end_of_story():
    game_result = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def flip_a_coin():
            @story.part()
            def tail_recursion(ctx):
                # TODO: add test for recursion
                # return flip_a_coin(message['user'], session)
                return EndOfStory({
                    'game_result': 'in progress'
                })

        @story.on('enter to the saloon')
        def enter_to_the_saloon():
            @story.part()
            async def start_a_game(ctx):
                return await flip_a_coin(session=ctx['session'], user=ctx['user'], )

            @story.part()
            def game_over(ctx):
                logger.debug('game_over')
                logger.debug(ctx)
                game_result.receive(story_context.get_user_data(ctx)['game_result'])

        await talk.pure_text('enter to the saloon')

        assert game_result.result() in ['loose', 'win', 'in progress']


@pytest.mark.asyncio
async def test_sync_end_of_story():
    part_1 = SimpleTrigger()
    part_2 = SimpleTrigger()
    part_3 = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def one_story():
            @story.part()
            def story_part_1(ctx):
                part_1.passed()

            @story.part()
            def story_part_2(ctx):
                part_2.passed()
                return EndOfStory('Break Point')

            @story.part()
            def story_part_3(ctx):
                part_3.passed()

        res = await one_story(session=talk.session, user=talk.user)

        assert part_1.is_triggered
        assert part_2.is_triggered
        assert not part_3.is_triggered
        assert res == 'Break Point'


@pytest.mark.asyncio
async def test_could_just_past_context_to_callable():
    session_storage = SimpleTrigger()
    user_storage = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def one_story():
            @story.part()
            def one_story_part(ctx):
                session_storage.receive(ctx['session']['user_id'])
                user_storage.receive(ctx['user'])

        @story.on('hi')
        def hi_story():
            @story.part()
            async def run_one_story(ctx):
                await one_story(ctx)

        await talk.pure_text('hi')

        assert session_storage.value == talk.session['user_id']
        assert user_storage.value == talk.user


# Loop & Callable Stories


@pytest.mark.asyncio
async def test_story_loop_inside_of_callable():
    inside_of_loop = SimpleTrigger()
    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def one_callable():
            @story.part()
            def story_init(ctx):
                pass

            @story.loop()
            def inner_loop():
                @story.on('jump')
                def jump_story():
                    @story.part()
                    def inner_job(ctx):
                        inside_of_loop.passed()

        # TODO: add wrapper for getting result context from callable to talk
        ctx = await one_callable(session=talk.session, user=talk.user,
                                 value='Hello World!')

        talk.session = ctx.message['session']

        await talk.pure_text('jump')

        assert inside_of_loop.is_passed()


@pytest.mark.asyncio
async def test_propagate_arguments_inside_of_story_loop():
    inner_job_receive = SimpleTrigger()
    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def one_callable():
            @story.part()
            def story_init(ctx):
                pass

            @story.loop()
            def inner_loop():
                @story.on('jump')
                def jump_story():
                    @story.part()
                    def inner_job(ctx):
                        inner_job_receive.receive(story_context.get_user_data(ctx)['value'])

        ctx = await one_callable(session=talk.session, user=talk.user,
                                 value='Hello World!')

        # TODO: add wrapper for talk
        talk.session = ctx.message['session']

        await talk.pure_text('jump')

        assert inner_job_receive.result() == 'Hello World!'


@pytest.mark.asyncio
async def test_exit_outside_of_callable_and_loop():
    exit_inside_callable_trigger = SimpleTrigger(0)
    in_progress_inside_callable_trigger = SimpleTrigger(0)
    in_progress_outside_callable_trigger = SimpleTrigger(0)
    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def one_callable():
            @story.part()
            def inner_init(ctx):
                pass

            @story.loop()
            def inner_loop():
                @story.on('in-progress')
                def in_progress():
                    @story.part()
                    def do_in_progress(ctx):
                        in_progress_inside_callable_trigger.inc()

                @story.on('exit')
                def exit():
                    @story.part()
                    def break_loop(ctx):
                        exit_inside_callable_trigger.inc()
                        return loop.BreakLoop()

        @story.on('start')
        def start_story():
            @story.part()
            async def call_callable(ctx):
                return await one_callable(ctx)

        @story.on('in-progress')
        def in_progress_outside_callable():
            @story.part()
            async def in_progress_outside_callable_store(ctx):
                in_progress_outside_callable_trigger.inc()

        await talk.pure_text('start')
        await talk.pure_text('in-progress')
        await talk.pure_text('exit')
        await talk.pure_text('in-progress')

        assert exit_inside_callable_trigger.result() == 1
        assert in_progress_inside_callable_trigger.result() == 1
        assert in_progress_outside_callable_trigger.result() == 1


@pytest.mark.asyncio
async def test_exit_outside_of_callable_and_loop_but_do_not_match_again():
    exit_outside_callable_trigger = SimpleTrigger(0)
    with answer.Talk() as talk:
        story = talk.story

        @story.callable()
        def one_callable():
            @story.part()
            def inner_init(ctx):
                pass

            @story.loop()
            def inner_loop():
                @story.on('in-progress')
                def in_progress():
                    @story.part()
                    def do_in_progress(ctx):
                        pass

                @story.on('exit')
                def exit():
                    @story.part()
                    def break_loop(ctx):
                        return loop.BreakLoop()

        @story.on('start')
        def start_story():
            @story.part()
            async def call_callable(ctx):
                return await one_callable(ctx)

        @story.on('exit')
        def exit_outside_callable():
            @story.part()
            async def exit_outside_callable_store(ctx):
                exit_outside_callable_trigger.inc()

        await talk.pure_text('start')
        await talk.pure_text('in-progress')
        await talk.pure_text('exit')

        assert exit_outside_callable_trigger.result() == 0
